#include <u.h>
#include <libc.h>
#include <ureg.h>
#include <draw.h>
#include <ctype.h>
#include "dat.h"
#include "fns.h"
#include "linux.h"
#include "tty.h"

enum {	/* OPOST state machine */
	STnorm		= 0,		/* vanilla text */
	STcr		= 1,		/* seen a \r */
	STesc		= 2,		/* seen an <esc> */
	STopen		= 3,		/* seen an '[' following an <esc> */

	Buftime 	= 100 * 1000 * 1000,
};



typedef struct Cons Cons;
struct Cons
{
	Ufile;
	void	*bufproc;
	char	buf[8192];
	char 	*w;
	int	state;
	int	inp;
	int	out;
	int	ctl;
	long	sem;
	int 	iscooked;
	int	isaconsole;
	struct linux_termios ts;
	Point 	chrsz;
	int pid;
};

static void*
bufproccons(Cons *cons)
{
	if(cons->bufproc == nil)
		cons->bufproc = newbufproc(0);
	return cons->bufproc;
}

static int
pollcons(Ufile *file, void *tab)
{
	Cons *cons = (Cons*)file;
	return pollbufproc(bufproccons(cons), cons, tab);
}

static int
readcons(Ufile *file, void *buf, int len, vlong)
{
	Cons *cons = (Cons*)file;
	int ret;

	if((cons->omode & O_NONBLOCK) || (cons->bufproc != nil)){
		ret = readbufproc(bufproccons(cons), buf, len, cons->inp, (cons->omode & O_NONBLOCK));
	} else {
		notifyme(1);
		ret = read(cons->inp, buf, len);
		notifyme(0);

		if(ret < 0)
			ret = mkerror();
	}
	return ret;
}

static int 
flushout(Cons *cons)
{
	int ret;

	if(cons->w - cons->buf == 0)
		return 0;

	ret = write(cons->out, cons->buf, cons->w - cons->buf);
	cons->w = cons->buf;
	if(ret < 0)
		return mkerror();
	return ret;
}

void
flusher(void *arg)
{
	Cons *cons = (Cons*)arg;

	while(1){
		semacquire(&cons->sem, 1);
		sleep(50);
		flushout(cons);
	}
}

static int
writecons(Ufile *file, void *buf, int len, vlong)
{
	int ret;
	char *p, c;
	Cons *cons = (Cons*)file;

	if(! cons->isaconsole || (cons->ts.c_oflag & OPOST) != OPOST){
		notifyme(1);
		ret = write(cons->out, buf, len);
		notifyme(0);
		return ret;
	}

	for(p = buf; p < (char *)buf+len; p++){
		c = *p;

		switch(cons->state){
		case STnorm:
			if(c == 0x1b){
				cons->state = STesc;
				continue;
			}
			if(c == '\r'){
				cons->state = STcr;
				continue;
			}
			break;
		case STesc:
			if(c == '['){
				cons->state = STopen; 
				continue;
			}
			*cons->w++ = 0x1b;
			cons->state = STnorm;
			break;
		case STopen:
			if(isdigit(c))
				continue;
			if(c == 'J' || c == 'm' || c == 'H'){
				cons->state = STnorm;
				continue;
			}
			*cons->w++ = 0x1b;
			*cons->w++ = '[';
			cons->state = STnorm;
			break;
		case STcr:
			if(c == '\r')
				continue;
			if(c != '\n'){
				*cons->w++ = '\n';
				cons->state = STnorm;
			}
		}

		if((cons->w - cons->buf) >= sizeof(cons->buf)-8){
			notifyme(1);
			if((ret = flushout(cons)) < 0){
				notifyme(0);
				return ret;
			}
			notifyme(0);
		}
		*cons->w++ = c;
	}

	if((cons->w - cons->buf) > 0)
		semrelease(&cons->sem, 1);
	return len;
}

static int
winsize(Cons *cons, void *arg)
{
	int n, fd;
	Winsize *ws;
	char *a[8], buf[128];

	ws = arg;
	fd = open("/dev/wctl", OREAD);
	if(fd == -1){
		print("/dev/wctl open failed  %r\n");
		return -EIO;
	}

	n = read(fd, buf, sizeof(buf));
		close(fd);
	if(n < 0){
		print("/dev/wctl read failed  %r\n");
		return -EIO;
	}

	buf[n-1] = 0;
	if((n = tokenize(buf, a, nelem(a))) < 4){
		print("/dev/wctl too few tokens (%d<4)\n", n);
		return -EIO;
	}
	/* Lifted from mc.c
	   window stucture:
		4 pixels left edge
		1 pixels gap
		12 pixels scrollbar
		4 pixels gap
		text
		4 pixels right edge

	  Steve: vertical edges - I guess at 4 pixels top and bottom.
	*/

	ws->px = atoi(a[2]) - atoi(a[0]) - (4+1+12+4+4);
	ws->py = atoi(a[3]) - atoi(a[1]) - (4+4);
	ws->row = ws->py / cons->chrsz.y;
	ws->col = ws->px / cons->chrsz.x;

	return 0;
}

static int
kitchen(Cons *cons, void *arg)
{
	char *ctlfile;
	int rc, cooked;
	struct linux_termios *tios;

	tios = arg;
	ctlfile = "/dev/consctl";
	cooked = ((tios->c_lflag & ECHO) == ECHO)? 1: 0;

	if(cooked == cons->iscooked)
		return 0;

	rc = -1;
	if(cooked){
		if(cons->ctl == -1){
			fprint(2, "kitchen: rawoff %s not open\n", ctlfile);
			return -1;
		}
		rc = write(cons->ctl, "rawoff", 6);
		close(cons->ctl);
		cons->ctl = -1;
	}
	else{
		if(cons->isaconsole){
			if(cons->ctl != -1){
				fprint(2, "kitchen: rawon %s already open\n", ctlfile);
				return -1;
			}
			if((cons->ctl = open(ctlfile, OWRITE)) == -1)
				fprint(2, "%s: %s cannot open - %r\n", argv0, ctlfile);
			if(cons->ctl != -1)
				rc = write(cons->ctl, "rawon", 5);
		}
	}

	if(rc < 0)
		fprint(2, "kitchen: %s write fail %r\n", ctlfile);
	else
		cons->iscooked = cooked;
	return 0;
}

static int
ioctlcons(Ufile *file, int cmd, void *arg)
{
	int r;
	Cons *cons = (Cons*)file;

	switch(cmd){
	default:
		if(debug)
			fprint(2, "ioctlcons: cmd=0x%x not supported\n", cmd);
		return -ENOSYS;
	case TCGETS:
		if(arg == nil)
			return -EINVAL;
		if(! cons->isaconsole)
			return -ENOTTY;
		memcpy(arg, &cons->ts, sizeof(cons->ts));
		break;
	case TCSETS:
	case TCSETSW:
	case TCSETSF:
		if(arg == nil)
			return -EINVAL;
		if(! cons->isaconsole)
			return -ENOTTY;
		if(kitchen(cons, arg) < 0){
			if(debug)
				fprint(2, "ioctlcons: kitchen() failed\n");
			return -EIO;
		}
		memcpy(&cons->ts, arg, sizeof(cons->ts));
		break;
	case TIOCGWINSZ:
		if(arg == nil)
			return -EINVAL;
		if(! cons->isaconsole)
			return -ENOTTY;
		return winsize(cons, arg);
	case FIONREAD:  	// AKA  TIOCINQ
		if(arg == nil)
			return -EINVAL;
		if((r = nreadablebufproc(bufproccons(cons))) < 0){
			*((int*)arg) = 0;
			return r;
		}
		*((int*)arg) = r;
		break;
	case TIOCGPGRP:
		if(arg == nil)
			return -EINVAL;
		if(! cons->isaconsole)
			return -ENOTTY;
		*(int *)arg = current->pgid;
		break;
	case TIOCSPGRP:
		if(arg == nil)
			return -EINVAL;
		if(! cons->isaconsole)
			return -ENOTTY;
		current->pgid = *(int *)arg;
		break;
	}
	return 0;
}

int
isaconsole(int fd)
{
	int m, n;
	char outpath[1024], *devcons;

	if(fd2path(fd, outpath, sizeof(outpath)) < 0)
		return 0;
	n = strlen(outpath);
	devcons = "/dev/cons";
	m = strlen(devcons);
	if(n < m)
		return 0;
	if(strcmp(outpath+(n-m), devcons) != 0)
		return 0;
	return 1;
}

static Point
fontsize(void)
{
	Font *f;
	Point sz;
	char *fontname;

	if((fontname = getenv("font")) == nil)
		return Pt(8, 12);

	if((f = openfont(nil, fontname)) == nil){
		fprint(2, "%s: %s cannot open - %r\n", argv0, fontname);
		free(fontname);
		return Pt(8, 12);
	}
	sz = stringsize(f, "0");
	freefont(f);
	free(fontname);
	return sz;
}

static int
opencons(char *path, int omode, int, Ufile **pf)
{
	Cons *cons;

	cons = mallocz(sizeof(Cons), 1);
	cons->ref = 1;
	cons->omode = omode;
	cons->dev = CONSDEV;
	cons->path = kstrdup(path);

	cons->inp = 0;
	cons->out = 1;
	cons->ctl = -1;
	cons->isaconsole = isaconsole(cons->inp);
	cons->chrsz = fontsize();
	cons->w = cons->buf;
	cons->iscooked = 1;

	cons->ts.c_cc[VINT] = 127;
	cons->ts.c_cc[VERASE] = 'H'-'@';
	cons->ts.c_cc[VEOF] = 'D'-'@';
	cons->ts.c_iflag = IUTF8;
	cons->ts.c_cflag = CREAD|B38400;
	cons->ts.c_lflag = ISIG|ICANON|ECHO;

	if(cons->isaconsole){
		cons->ts.c_oflag = OPOST;
		cons->pid = procfork(flusher, cons, 0);
	}
	
	*pf = cons;
	return 0;
}

static int
closecons(Ufile *file)
{
	Cons *cons = (Cons*)file;

	if(cons->ctl != -1)
		close(cons->ctl);
	postnote(PNPROC, cons->pid, "die yankee!");
	flushout(cons);
	freebufproc(cons->bufproc);
	close(cons->ctl);
	return 0;
}

static int
statcons(char *, int, Ustat *s)
{
	s->mode = 0666 | S_IFCHR;
	s->uid = current->uid;
	s->gid = current->gid;
	s->size = 0;
	s->ino = 1;
	s->atime = s->mtime = s->ctime = boottime/1000000000LL;
	s->dev = CONSDEV;
	s->rdev = makedev(5, 1);
	return 0;
}

static int
fstatcons(Ufile *f, Ustat *s)
{
	return fsstat(f->path, 0, s);
};

static Udev consdev = 
{
	.name = "console",
	.open = opencons,
	.read = readcons,
	.write = writecons,
	.poll = pollcons,
	.close = closecons,
	.ioctl = ioctlcons,
	.fstat = fstatcons,
	.stat = statcons,
};

void
consdevinit(void)
{
	devtab[CONSDEV] = &consdev;
	fsmount(&consdev, "/dev/console");
}
