#include <u.h>
#include <libc.h>
#include <ureg.h>
#include <mp.h>
#include <libsec.h>
#include <draw.h>

#include "dat.h"
#include "fns.h"
#include "linux.h"

enum {
	HEIGHT = 303,			/* 12 inches */
	PHYSADDR = 0x1000000,
	FB_TYPE_PACKED_PIXELS = 0,
	FB_VISUAL_TRUECOLOR = 2,

};

struct fb_bitfield {
	uint offset;	       /* beginning of bitfield        */
	uint length;	       /* length of bitfield           */
	uint msb_right;	       /* != 0 : Most significant bit is right */
};

struct fb_var_screeninfo {
	uint xres;	       /* visible resolution           */
	uint yres;
	uint xres_virtual;     /* virtual resolution           */
	uint yres_virtual;
	uint xoffset;	       /* offset from virtual to visible */
	uint yoffset;	       /* resolution                   */

	uint bits_per_pixel;   /* guess what                   */
	uint grayscale;	       /* != 0 Graylevels instead of colors */

	struct fb_bitfield red;	/* bitfield in fb mem if true color, */
	struct fb_bitfield green;	/* else only length is significant */
	struct fb_bitfield blue;
	struct fb_bitfield transp;	/* transparency                 */

	uint nonstd;	       /* != 0 Non standard pixel format */

	uint activate;	       /* see FB_ACTIVATE_*            */

	uint height;	       /* height of picture in mm    */
	uint width;	       /* width of picture in mm     */

	uint accel_flags;      /* (OBSOLETE) see fb_info.flags */

	/* Timing: All values in pixclocks, except pixclock (of course) */
	uint pixclock;	       /* pixel clock in ps (pico seconds) */
	uint left_margin;      /* time from sync to picture    */
	uint right_margin;     /* time from picture to sync    */
	uint upper_margin;     /* time from sync to picture    */
	uint lower_margin;
	uint hsync_len;	       /* length of horizontal sync    */
	uint vsync_len;	       /* length of vertical sync      */
	uint sync;	       /* see FB_SYNC_*                */
	uint vmode;	       /* see FB_VMODE_*               */
	uint rotate;	       /* angle we rotate counter clockwise */
	uint reserved[5];      /* Reserved for future compatibility */
};

struct fb_fix_screeninfo {
	char id[16];	       /* identification string eg "TT Builtin" */
	ulong smem_start;      /* Start of frame buffer mem (physical address) */
	uint smem_len;	       /* Length of frame buffer mem */
	uint type;	       /* see FB_TYPE_*                */
	uint type_aux;	       /* Interleave for interleaved Planes */
	uint visual;	       /* see FB_VISUAL_*              */
	ushort xpanstep;       /* zero if no hardware panning  */
	ushort ypanstep;       /* zero if no hardware panning  */
	ushort ywrapstep;      /* zero if no hardware ywrap    */
	uint line_length;      /* length of a line in bytes    */
	unsigned long mmio_start; /* Start of Memory Mapped I/O (physical address) */
	uint mmio_len;	       /* Length of Memory Mapped I/O  */
	uint accel;	       /* Indicate to driver which specific chip/card we have  */
	ushort reserved[3];    /* Reserved for future compatibility */
};

struct Framebuffer {
	Image *i;
	uchar *buf;
	int siz;
	uvlong sum;
	int pid;
};

static uvlong
sum(void *buf, void *end) /* NB: assumes args are on vlong boundries */
{
	uvlong *p, n;

	n = 0;
	p = buf;
	while(p < end)
		n += *p++;
	return n;
}

static void
flushproc(Framebuffer *fb)
{
	uvlong n, m;

fprint(2, "child2=%d fb=%p\n", getpid(), fb);
	/* FIXME: need real locking here one day, but this will do for now */
	while(1){
fprint(2, "["); 
		sleep(100);
fprint(2, "]"); 
		n = sum(fb->buf, fb->buf + fb->siz);
		if(n == m)
			continue;
		m = n;
		loadimage(fb->i, fb->i->r, fb->buf, fb->siz);
		draw(screen, screen->r, fb->i, nil, fb->i->r.min);
		flushimage(display, 1);
	}
}

Framebuffer *
fb_open(void **fbmem, ulong *fbsiz)
{
	Image *i;
	Framebuffer *fb;

	if(initdraw(nil, nil, nil) < 0){
		kprint("jpg: initdraw failed: %r\n");
		return nil;
	}

	i = allocimage(display, screen->clipr, screen->chan, 0, DWhite);
	if(i == nil){
		kprint("jpg: initdraw failed: %r\n");
		return nil;
	}

	fb = kmallocz(sizeof(Framebuffer), 1);
	fb->i = i;
	fb->siz = pagealign(bytesperline(fb->i->r, fb->i->depth) * Dy(fb->i->r));
	fb->buf = (uchar *)sys_mmap(PHYSADDR, fb->siz, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS|MAP_FIXED, -1, 0);

fprint(2, "parent1=%d fb=%p\n", getpid(), fb);
	fb->pid = rfork(RFPROC|RFMEM);
	switch(fb->pid){
	case -1:	/* failed */
fprint(2, "parent2=%d fb=%p\n", getpid(), fb);
		sysfatal("fork failed %r\n");
		break;
	case 0:		/* child */
fprint(2, "child1=%d fb=%p\n", getpid(), fb);
		flushproc(fb);
		break;
	default:	/* parent */
fprint(2, "parent3=%d fb=%p\n", getpid(), fb);
		break;
	}

fprint(2, "parent4=%d fb=%p\n", getpid(), fb);

	*fbmem = fb->buf;		/* held in file struct so sys_linux_mmap() can find it */
	*fbsiz = fb->siz;
	return fb;
}

int
fb_read(Framebuffer *fb, void *buf, int len, vlong off)
{
	if(fb->siz - off < len)
		len = fb->siz - off;
	if(len < 0)
		return 0;
	memcpy(buf, fb->buf + off, len);
	return len;
}

int
fb_write(Framebuffer *fb, void *buf, int len, vlong off)
{
	if(fb->siz - off < len)
		len = fb->siz - off;
	if(len < 0)
		return 0;
	memcpy(fb->buf + off, buf, len);
	return len;
}

int
fb_getvarinfo(Framebuffer *fb, void *arg)
{
	struct fb_var_screeninfo vs;
	
	memset(&vs, 0, sizeof(vs));
	vs.xres = Dx(fb->i->r);
	vs.yres = Dy(fb->i->r);

	vs.xres_virtual = Dx(fb->i->clipr);
	vs.yres_virtual = Dy(fb->i->clipr);

	vs.bits_per_pixel = fb->i->depth;
	switch(fb->i->depth){
	case 16:
		vs.red.length = 5;
		vs.green.length = 6;
		vs.green.offset = 5;
		vs.blue.length = 5;
		vs.blue.offset = 11;
		break;
	case 24:
		vs.red.length = 8;
		vs.green.length = 8;
		vs.green.offset = 8;
		vs.blue.length = 8;
		vs.blue.offset = 16;
		break;
	case 32:
		vs.red.length = 8;
		vs.green.length = 8;
		vs.green.offset = 8;
		vs.blue.length = 8;
		vs.blue.offset = 16;
		vs.transp.length = 8;
		vs.transp.offset = 24;
		break;
	default:
		kprint("%d unsupported display depth\n", fb->i->depth);
		return -ENOSYS;
	}

	vs.height = HEIGHT;		/* physical size in mm */
	vs.width = (vs.xres * vs.height) / vs.yres;
	memcpy(arg, &vs, sizeof(vs));
	return 0;
}

int
fb_putvarinfo(Framebuffer *fb, void *arg)
{
	int err;
	struct fb_var_screeninfo vsi;
	
	if((err = fb_getvarinfo(fb, &vsi)) != 0)
		return err;
	if(memcmp(&vsi, arg, sizeof(vsi)) != 0)
		return -EIO;
	return 0;
}

int
fb_getfixinfo(Framebuffer *fb, char *arg)
{
	struct fb_fix_screeninfo fs;

	memset(&fs, 0, sizeof(fs));
	snprint(fs.id, sizeof(fs.id), "Plan9");

	fs.smem_start = (ulong)fb->buf;
	fs.smem_len = fb->siz;
	fs.type = FB_TYPE_PACKED_PIXELS;
	fs.visual = FB_VISUAL_TRUECOLOR;
//	fs.ypanstep = 1;
	fs.line_length = bytesperline(fb->i->r, fb->i->depth);

	memcpy(arg, &fs, sizeof(fs));
	return 0;
}

int
fb_close(Framebuffer *fb)
{
	postnote(PNPROC, fb->pid, "die yankee pig");
	sleep(10);	/* let flusher die */

	freeimage(fb->i);
/* FIXME - do we need to free the buffer of is that done as part of the process shutdown for us ? */
	free(fb);
	return 0;
}
