#include <u.h>
#include <libc.h>
#include <ureg.h>
#include "dat.h"
#include "fns.h"
#include "linux.h"

typedef struct Mount Mount;

struct Mount
{
	Mount	*next;
	Udev		*dev;
	int		npath;
	char		path[];
};

static Mount *mtab;

void
fsmount(Udev *dev, char *path)
{
	Mount *m, **p;
	int n;

	if(dev == nil)
		return;

	n = strlen(path);
	m = kmalloc(sizeof(*m) + n + 1);
	m->dev = dev;
	m->next = nil;
	m->npath = n;
	strcpy(m->path, path);

	for(p=&mtab;;p=&((*p)->next)){
		Mount *x;

		if(x = *p){
			if(m->npath < x->npath)
				continue;
			if(m->npath == x->npath){
				if(strcmp(m->path, x->path) < 0)
					continue;
			}
		}
		m->next = *p;
		*p = m;
		break;
	}
}

char *
mounts(void)
{
	Fmt f;
	Mount *m;

	fmtstrinit(&f);
	for(m = mtab; m ;m = m->next)
		fmtprint(&f, "%-8s %-16s %-8s	defaults	0	0\n",
			m->dev->name, (*m->path == 0)? "/": m->path, m->dev->name);
	return fmtstrflush(&f);
}

static int
open_devdir(char *path, int omode, int, Ufile **pf)
{
	Ufile *f;

	if(strcmp(path, "/dev") != 0)
		return -ENOENT;
	f = kmallocz(sizeof(Ufile), 1);
	f->ref = 1;
	f->omode = omode;
	f->path = kstrdup(path);
	f->fd = -1;
	f->dev = DEVDIR;
	*pf = f;
	return 0;
}

static int
readdir_devdir(Ufile *, Udirent **pd)
{
	int n;
	Mount *m;

	n = 0;
	for(m = mtab; m ;m = m->next){
		if(strncmp(m->path, "/dev/", 5) != 0)
			continue;
		if((*pd = newdirent(m->path, S_IFDIR|0755, mkino(DEVDIR, n))) == nil)
			break;
		pd = &((*pd)->next);
		n++;
	}
	return n;
}


static int
close_devdir(Ufile *)
{
	return 0;
}

static int
stat_devdir(char *path, int link, Ustat *s)
{
	USED(link, path);
	if(strcmp(path, "/dev") != 0)
		return -ENOENT;
	s->mode = S_IFDIR|0755;
	s->uid = 0;
	s->gid = 0;
	s->size = 0;
	s->ino = mkino(99, 0);
	s->dev = DEVDIR;
	s->rdev = 0;
	s->atime = s->mtime = s->ctime = boottime/1000000000LL;
	return 0;
}

static int
fstat_devdir(Ufile *f, Ustat *s)
{
	return fsstat(f->path, 0, s);
};

static Udev devdir =
{
	.name = "devdir",
	.open = open_devdir,
	.readdir = readdir_devdir,
	.close = close_devdir,
	.stat = stat_devdir,
	.fstat = fstat_devdir,
};

void
devdirinit(void)
{
	devtab[DEVDIR] = &devdir;
	fsmount(&devdir, "/dev");
}

ulong
hashpath(char *s)
{
	ulong h;
	for(h=0; *s; s++)
		h = (h * 13) + (*s - 'a');
	return h;
}

char*
basepath(char *p, char **ps)
{
	char *x, *s;
	int n;

	if(s = strrchr(p, '/')){
		if(s[1] != 0){
			if(ps)
				*ps = kstrdup(s+1);
			if((n = s - p) == 0)
				n = 1;
			x = kmalloc(n+1);
			memmove(x, p, n);
			x[n] = 0;
			return x;
		}
	}
	if(ps)
		*ps = nil;
	return nil;
}

char*
allocpath(char *base, char *prefix, char *name)
{
	char *p, *s;
	int n, m, k;

	n = strlen(base);
	m = strlen(name);
	k = prefix ? strlen(prefix) : 0;
	p = s = kmalloc(n+m+k+2);
	memmove(p, base, n);
	p += n;
	if(m || k)
		*p++ = '/';
	if(k){
		memmove(p, prefix, k);
		p += k;
	}
	memmove(p, name, m+1);
	return s;
}

char*
fullpath(char *base, char *name)
{
	char *s;

	if(*name == '/' || *name == '#'){
		s = kstrdup(name);
	} else if(base) {
		s = allocpath(base, nil, name);
	} else {
		s = nil;
	}
	if(s)
		cleanname(s);
	return s;
}


int
fsfullpath(char **path, int check)
{
	int err;
	char *s, *newpath;

	if(**path == 0)
		return -EINVAL;

	if((newpath = fullpath(current->cwd, *path)) == nil)
		return -EFAULT;
	
	if(current->root){
		s = allocpath(current->root, nil, newpath+1);
		free(newpath);
		newpath = s;
	}

	if((err = lookcache(&newpath, check)) < 0){
		if(debug)
			if(access(newpath, 0) != -1)
				print("cache out of sync: %s\n", newpath);
		free(newpath);
		return err;
	}
	*path = newpath;
	return 0;
}

static Mount*
path2mount(char *path)
{
	Mount *m;

	for(m=mtab; m; m=m->next){
		if(strncmp(path, m->path, m->npath) == 0){
			switch(path[m->npath]){
			case '\0':
			case '/':
				return m;
			}
		}
	}
	return nil;
}

static Udev*
path2dev(char *path)
{
	Mount *m;

	if(m = path2mount(path))
		return m->dev;
	return nil;
}

int
sys_getcwd(char *buf, int len)
{
	int n;
	char *cwd;

	cwd = current->cwd;
	n = strlen(cwd)+1;
	if(n > len)
		return -ERANGE;
	memmove(buf, cwd, n);
	return n;
}

int
fsaccess(char *path, int mode)
{
	int err;
	Udev *dev;

	err = -ENOENT;
	if(dev = path2dev(path)){
		err = 0;
		if(dev->access)
			err = dev->access(path, mode);
	}

	return err;
}

int
do_faccess(char *path, int mode, int flags)
{
	int err;

	if((err = fsfullpath(&path, Pfull)) < 0)
		return err;
again:	if((err = fsaccess(path, mode)) < 0)
		if((flags & AT_SYMLINK_NOFOLLOW) == 0)
			if((err = retry(err, &path, Pfull)) == 0)
				goto again;
	free(path);
	return err;
}

int
sys_access(char *path, int mode)
{
	return do_faccess(path, mode, 0);
}

int
fsopen(char *path, int mode, int perm, Ufile **pf)
{
	int err;
	Udev *dev;

	*pf = nil;
	err = -ENOENT;
	if((dev = path2dev(path)) && dev->open)
		err = dev->open(path, mode, perm, pf);
	return err;
}

int
sys_open(char *path, int omode, int perm)
{
	Ufile *file;
	int err, check;

	/*
	 * NB: open with O_CREAT means the file is created if it doesn't exist
	 * but is opened otherwise. however if it does exists it may be a symlink
	 * we have to follow (e.g. /dev/tty) so we use Poptional mode for this
	 */
	check = ((omode & O_CREAT) == O_CREAT)? Poptional: Pfull;

	if((err = fsfullpath(&path, check)) < 0)
		return err;
again: if((err = fsopen(path, omode, perm, &file)) < 0)
		if((err = retry(err, &path, check)) == 0)
			goto again;
	if(err == 0){
		if(check == Poptional)
			flushcache(path);
		err = newfd(file, FD_CLOEXEC);
	}
	free(path);

	return err;
}

int
sys_creat(char *path, int perm)
{
	int err;
	Ufile *f;

	if((err = fsfullpath(&path, Ppartial)) < 0)
		return err;
again:	if((err = fsopen(path, O_CREAT|O_TRUNC, perm, &f)) < 0)
		if((err = retry(err, &path, Ppartial)) == 0)
			goto again;
	if(err == 0){
		flushcache(path);
		err = newfd(f, FD_CLOEXEC);
	}
	free(path);

	return err;
}

int
fsstat(char *path, int link, Ustat *ps)
{
	int err;
	Udev *dev;

	err = -EPERM;
	if((dev = path2dev(path)) && dev->stat){
		memset(ps, 0, sizeof(Ustat));
		err = dev->stat(path, link, ps);
	}
	return err;
}

int
sys_chdir(char *path)
{
	int err;
	Ufile *f;

	if((err = fsfullpath(&path, Pfull)) < 0)
		return err;
again:	if((err = fsopen(path, O_RDONLY, 0, &f)) < 0)
		if((err = retry(err, &path, Pfull)) == 0)
			goto again;	
	free(path);
	if(err == 0){
		err = chdirfile(f);
		putfile(f);
	}
	return err;
}

int
sys_chroot(char *path)
{
	Ufile *f;
	Ustat s;
	int err;

	f = nil;
	if((err = fsfullpath(&path, Pfull)) < 0)
		return err;
again:	if((err = fsopen(path, O_RDONLY, 0, &f)) < 0)
		if((err = retry(err, &path, Pfull)) == 0)
			goto again;
	if(err < 0)
		goto out;

	err = -ENOTDIR;
	if(f->path == nil)
		goto out;
	if(devtab[f->dev]->fstat == nil)
		goto out;
	if((err = devtab[f->dev]->fstat(f, &s)) < 0)
		goto out;
	err = -ENOTDIR;
	if((s.mode & ~0777) != S_IFDIR)
		goto out;
	err = 0;
	free(current->root);
	if(strcmp(f->path, "/") == 0){
		current->root = nil;
	} else {
		current->root = kstrdup(f->path);
	}
out:
	putfile(f);
	free(path);
	return err;
}

int
fschown(char *path, int uid, int gid, int link)
{
	int err;
	Udev *dev;

	err = -EPERM;
	if((dev = path2dev(path)) && dev->chown)
		err = dev->chown(path, uid, gid, link);
	return err;
}

int
sys_chown(char *path, int uid, int gid)
{
	int err;

	if((err = fsfullpath(&path, Pfull)) < 0)
		return err;
again: if((err = fschown(path, uid, gid, 0)) < 0)
		if((err = retry(err, &path, Pfull)) == 0)
			goto again;
	free(path);

	return err;
}

int
sys_lchown(char *path, int uid, int gid)
{
	int err;

	/*
	 * pass Ppartial to fsfullpath() and retry() so we don't try
	 * to dereference the link we are trying to chown
	 */
	if((err = fsfullpath(&path, Ppartial)) < 0)
		return err;
again: if((err = fschown(path, uid, gid, 1)) < 0)
		if((err = retry(err, &path, Ppartial)) == 0)
			goto again;
	free(path);

	return err;
}

int
fsreadlink(char *path, char *buf, int len)
{
	int err;
	Udev *dev;

	err = -EPERM;
	if((dev = path2dev(path)) && dev->readlink)
		err = dev->readlink(path, buf, len);

	return err;
}

int
sys_readlink(char *path, char *buf, int len)
{
	int err;

	/*
	 * pass Ppartial to fsfullpath() and retry() so we don't try
	 * to dereference the link we are trying to read
	 */
	if((err = fsfullpath(&path, Ppartial)) < 0)
		return err;
again:	if((err = fsreadlink(path, buf, len)) < 0)
		if((err = retry(err, &path, Ppartial)) == 0)
			goto again;
	free(path);
	return err;
}

int
fsrename(char *old, char *new)
{
	int err;
	Udev *dev;

	err = -EPERM;
	if((dev = path2dev(old)) && dev->rename){
		err = -EXDEV;
		if(dev == path2dev(new))
			err = dev->rename(old, new);
	}

	return err;
}


int
sys_rename(char *from, char *to)
{
	int err;

	if(*from == 0 || *to == 0)
		return -EINVAL;

	/*
	 * NB: don't expand the full link as renames of 
	 * symlinks should rename the link and not the target
	 */
	if((err = fsfullpath(&from, Ppartial)) < 0)
		return err;
	if((err = fsfullpath(&to, Ppartial)) < 0){
		free(from);
		return err;
	}

again:	if((err = fsrename(from, to)) < 0)
		if((err = retry(err, &from, Ppartial)) == 0)
			if((err = retry(err, &to, Ppartial)) == 0)
				goto again;

	if(err == 0){
		flushcache(from);
		flushcache(to);
	}
	free(to);
	return err;
}

int
fsmkdir(char *path, int mode)
{
	int err;
	Udev *dev;

	err = -EPERM;
	if((dev = path2dev(path)) && dev->mkdir)
		err = dev->mkdir(path, mode);

	return err;
}

int
sys_mkdir(char *path, int mode)
{
	int err;

	if((err = fsfullpath(&path, Ppartial)) < 0)
		return err;
again: if((err = fsmkdir(path, mode)) < 0)
		if((err = retry(err, &path, Ppartial)) == 0)
			goto again;
	if(err == 0)
		flushcache(path);
	free(path);

	return err;
}

int
fsutime(char *path, int atime, int mtime)
{
	int err;
	Udev *dev;

	err = -EPERM;
	if((dev = path2dev(path)) && dev->utime)
		err = dev->utime(path, atime, mtime);

	return err;
}

struct linux_utimbuf
{
	long	atime;
	long	mtime;
};

int
sys_utime(char *path, void *times)
{
	int err;
	long atime, mtime;
	struct linux_utimbuf *t = times;

	if(t){
		atime = t->atime;
		mtime = t->mtime;
	}
	else{
		atime = mtime = time(0);
	}

	if((err = fsfullpath(&path, Pfull)) < 0)
		return err;
again:	if((err = fsutime(path, atime, mtime)) < 0)
		if((err = retry(err, &path, Pfull)) == 0)
			goto again;

	free(path);

	return err;
}

int
sys_utimes(char *path, void *tvp)
{
	int err;
	long atime, mtime;
	struct linux_timeval *t = tvp;

	if(t){
		atime = t[0].tv_sec;
		mtime = t[1].tv_sec;
	}
	else{
		atime = mtime = time(0);
	}

	if((err = fsfullpath(&path, Pfull)) < 0)
		return err;
again:	if((err = fsutime(path, atime, mtime)) < 0)
		if((err = retry(err, &path, Pfull)) == 0)
			goto again;

	free(path);

	return err;
}

int
fschmod(char *path, int mode, int link)
{
	int err;
	Udev *dev;

	err = -EPERM;
	if((dev = path2dev(path)) && dev->chmod)
		err = dev->chmod(path, mode, link);

	return err;
}

int
sys_chmod(char *path, int mode)
{
	int err;

	if((err = fsfullpath(&path, Pfull)) < 0)
		return err;
again:	if((err = fschmod(path, mode, 1)) < 0)
		if((err = retry(err, &path, Pfull)) == 0)
			goto again;
	free(path);

	return err;
}

int
sys_lchmod(char *path, int mode)
{
	int err;

	/*
	 * NB: pass in Ppartial as we don't want to follow
	 * the link we are trying to chmod
	 */
	if((err = fsfullpath(&path, Ppartial)) < 0)
		return err;
again:	if((err = fschmod(path, mode, 0)) < 0)
		if((err = retry(err, &path, Ppartial)) == 0)
			goto again;
	free(path);

	return err;
}

int
fstruncate(char *path, vlong size)
{
	int err;
	Udev *dev;

	err = -EPERM;
	if((dev = path2dev(path)) && dev->truncate)
		err = dev->truncate(path, size);

	return err;
}

int
sys_truncate(char *path, ulong size)
{
	int err;

	if((err = fsfullpath(&path, Pfull)) < 0)
		return err;
again:	if((err = fstruncate(path, size)) < 0)
		if((err = retry(err, &path, Pfull)) == 0)
			goto again;
	free(path);

	return err;
}

int
fsunlink(char *path, int rmdir)
{
	int err;
	Udev *dev;

	err = -EPERM;
	if((dev = path2dev(path)) && dev->unlink)
		err = dev->unlink(path, rmdir);

	return err;
}

int
sys_unlink(char *path)
{
	int err;

	/*
	 * NB: don't expand the full symlink as unlink()'s of 
	 * symlinks should unlink() the symlink and not the target
	 */
	if((err = fsfullpath(&path, Ppartial)) < 0)
		return err;
again:	if((err = fsunlink(path, 0)) < 0)
		if((err = retry(err, &path, Ppartial)) == 0)
			goto again;
	if(err == 0)
		flushcache(path);
	free(path);

	return err;
}

int
sys_rmdir(char *path)
{
	int err;

	if((err = fsfullpath(&path, Pfull)) < 0)
		return err;
again:	if((err = fsunlink(path, 1)) < 0)
		if((err = retry(err, &path, Pfull)) == 0)
			goto again;
	if(err == 0)
		flushcache(path);
	free(path);

	return err;
}

int
fslink(char *old, char *new, int sym)
{
	int err;
	Udev *dev;

	err = -EPERM;
	if((dev = path2dev(new)) && dev->link){
		err = -EXDEV;
		if(sym || dev == path2dev(old))
			err = dev->link(old, new, sym);
	}

	return err;
}

int
sys_link(char *old, char *new)
{
	int err;

	if((err = fsfullpath(&old, Pfull)) < 0)
		return err;
	if((err = fsfullpath(&new, Ppartial)) < 0){
		free(old);
		return err;
	}

again:	if((err = fslink(old, new, 0)) < 0)
		if((err = retry(err, &old, Pfull)) == 0)
			if((err = retry(err, &new, Ppartial)) == 0)
				goto again;
	if(err == 0)
		flushcache(new);
	free(old);
	free(new);

	return err;
}

int
sys_symlink(char *old, char *new)
{
	int err;

	if((err = fsfullpath(&new, Ppartial)) < 0)
		return err;
again:	if((err = fslink(old, new, 1)) < 0)
		if((err = retry(err, &new, Ppartial)) == 0)
			goto again;
	if(err == 0)
		flushcache(new);
	return err;
}

int
fsmknod(char *path, int mode, int devnum)
{
	Udev *dev;

	if((dev = path2dev(path)) == nil || dev->mknod == nil)
		return -EPERM;
	return dev->mknod(path, mode, devnum);
}

int
sys_mknod(char *path, int mode, int dev)
{
	int err;

	if((err = fsfullpath(&path, Ppartial)) < 0)
		return err;
again:	if((err = fsmknod(path, mode, dev)) < 0)
		if((err = retry(err, &path, Pfull)) == 0)
			goto again;
	if(err == 0)
		flushcache(path);
	free(path);

	return err;
}


/*******************
 * atXXX() funcs
 */
static char *
atpath(int fd, char *path)
{
	char *s;
	Ufile *f;

	if((f = fdgetfile(fd)) == nil)
		return nil;
	if(path == nil)
		s = ksmprint("%s", f->path);
	else
		s = ksmprint("%s/%s", f->path, path);
	putfile(f);
	return s;
	
}

int
sys_openat(int fd, char *path, int flags, int mode)
{
	int err;

	if(fd == AT_FDCWD || *path == '/')
		return sys_open(path, flags, mode);

	if((path = atpath(fd, path)) == nil)
		return -EBADF;
	err = sys_open(path, flags, mode);
	free(path);
	return err;
}

int
sys_unlinkat(int fd, char *path, int flags)
{
	int err;

	if(fd == AT_FDCWD || *path == '/')
		if(flags & AT_REMOVEDIR)
			return sys_rmdir(path);
		else
			return sys_unlink(path);

	if((path = atpath(fd, path)) == nil)
		return -EBADF;
	if(flags & AT_REMOVEDIR)
		err = sys_rmdir(path);
	else
		err = sys_unlink(path);
	free(path);
	return err;
}

int
sys_mkdirat(int fd, char *path, int mode)
{
	int err;

	if(fd == AT_FDCWD || *path == '/')
		return sys_mkdir(path, mode);

	if((path = atpath(fd, path)) == nil)
		return -EBADF;
	err = sys_mkdir(path, mode);
	free(path);
	return err;
}

int
sys_renameat(int oldfd, char *oldpath, char *newpath)
{
	int err;

	if(oldfd == AT_FDCWD || *oldpath == '/')
		return sys_rename(oldpath, newpath);

	if((oldpath = atpath(oldfd, oldpath)) == nil)
		return -EBADF;
	err = sys_rename(oldpath, newpath);
	free(oldpath);
	return err;
}

int
sys_linkat(int oldfd, char *oldpath, int newfd, char *newpath, int flags)
{
	int err;
	char *n, *o;

	n = nil;
	o = nil;
	if(flags != 0)		/* currently unused, must be zero */
		return -EINVAL;

	if(oldfd != AT_FDCWD && *oldpath != '/')
		if((oldpath = o = atpath(oldfd, oldpath)) == nil)
			return -EBADF;
	if(newfd != AT_FDCWD && *newpath != '/')
		if((newpath = n = atpath(newfd, newpath)) == nil)
			return -EBADF;
	err = sys_link(oldpath, newpath);
	free(o);
	free(n);
	return err;
}

int
sys_futimesat(int fd, char *path, void *times)
{
	int err;

	if(fd == AT_FDCWD || *path == '/')
		return sys_utimes(path, times);

	if((path = atpath(fd, path)) == nil)
		return -EBADF;
	err = sys_utimes(path, times);
	free(path);
	return err;
}

int
sys_fstatat64(int fd, char *path, void *sb, int flags)
{
	int err;

	if(fd == AT_FDCWD || *path == '/')
		if(flags & AT_SYMLINK_NOFOLLOW)
			return sys_linux_lstat64(path, sb);
		else
			return sys_linux_stat64(path, sb);

	if((path = atpath(fd, path)) == nil)
		return -EBADF;
	if(flags & AT_SYMLINK_NOFOLLOW)
		err = sys_linux_lstat64(path, sb);
	else
		err = sys_linux_stat64(path, sb);
	free(path);
	return err;
}

int
sys_readlinkat(int fd, char *path, void *buf, int len)
{
	int err;

	if(fd == AT_FDCWD || *path == '/')
		return sys_readlink(path, buf, len);

	if((path = atpath(fd, path)) == nil)
		return -EBADF;
	err = sys_readlink(path, buf, len);
	free(path);
	return err;
}

int
sys_fchownat(int fd, char *path, int uid, int gid, int flags)
{
	int err;

	if(fd == AT_FDCWD || *path == '/')
		if(flags & AT_SYMLINK_NOFOLLOW)
			return sys_chown(path, uid, gid);
		else
			return sys_lchown(path, uid, gid);

	if((path = atpath(fd, path)) == nil)
		return -EBADF;
	if(flags & AT_SYMLINK_NOFOLLOW)
		err = sys_chown(path, uid, gid);
	else
		err = sys_lchown(path, uid, gid);
	free(path);
	return err;
}

int
sys_fchmodat(int fd, char *path, int mode, int flags)
{
	int err;

	if(fd == AT_FDCWD || *path == '/')
		if(flags & AT_SYMLINK_NOFOLLOW)
			return sys_chmod(path, mode);
		else
			return sys_lchmod(path, mode);

	if((path = atpath(fd, path)) == nil)
		return -EBADF;
	if(flags & AT_SYMLINK_NOFOLLOW)
		err = sys_chmod(path, mode);
	else
		err = sys_lchmod(path, mode);
	free(path);
	return err;
}

int
sys_symlinkat(char *old, int fd, char *new)
{
	int err;

	if(fd == AT_FDCWD || *new == '/')
		return sys_symlink(old, new);

	if((new = atpath(fd, new)) == nil)
		return -EBADF;
	err = sys_symlink(old, new);
	free(new);
	return err;
}

int
sys_faccessat(int fd, char *path, int mode, int flags)
{
	int err;

	if(fd == AT_FDCWD || *path == '/')
		return do_faccess(path, mode, flags);

	if((path = atpath(fd, path)) == nil)
		return -EBADF;
	err = do_faccess(path, mode, flags);
	free(path);
	return err;
}

