#include <u.h>
#include <libc.h>
#include <ureg.h>
#include <tos.h>
#include "dat.h"
#include "fns.h"
#include "linux.h"
#include "ioctlnames.h"

struct Linuxcall {
	char	*name;
	char 	*args;
	void	*func;
	vlong	calls;
	vlong	min[3];
	vlong	max[3];
	vlong	tot[3];
};

enum {
	User = 0,
	Kern,
	Real
};

static int
trampoline(Ureg *u, void *func)
{
	return ((int (*)(int, int, int, int, int, int))func)(u->bx, u->cx, u->dx, u->si, u->di, u->bp);
}

static int
sys_nop(void)
{
	return 0;
}

static int
sys_nosys(void)
{
	trace("syscall %s() - systemcall not implemented", current->syscall);
	kprint("%s: %s() - systemcall not implemented\n", current->comm, current->syscall);

//	if(debug)
//		abort();
	return -ENOSYS;
}


#include "linuxcalltab.out"

static void
linuxret(int errno)
{
	vlong n;
	Uproc *p;
	Ureg *u;
	Linuxcall *c;
	uvlong leave;


	p = current;
	u = p->ureg;
	c = p->calltab;		/* c == nil on child return from clone() */

	if(errno > 0)
		trace("%d", errno);
	else
		trace("%E", errno);

	if(c){
		cycles(&leave);

		n = leave - p->enter;
		c->tot[Real] += n;
		if(n > c->max[Real])
			c->max[Real] = n;
		if(c->min[0] == 0 || n < c->min[Real])
			c->min[Real] = n;
	
		if(_tos->pcycles){
			n = _tos->pcycles - p->pcycles;
n = -n;
			c->tot[User] += n;
			if(n > c->max[User])
				c->max[User] = n;
			if(c->min[0] == 0 || n < c->min[User])
				c->min[User] = n;
		}
	
		if(_tos->kcycles){
			n = _tos->kcycles - p->kcycles;
			c->tot[Kern] += n;
			if(n > c->max[Kern])
				c->max[Kern] = n;
			if(c->min[1] == 0 || n < c->min[Kern])
				c->min[Kern] = n;
		}
	}

	if(errno == -ERESTART){
		p->restart->syscall = p->syscall;
		return;
	}

	u->ax = (ulong)errno;
	u->pc += 2;			/* step over the int80 */
	p->restart->syscall = nil;
	p->syscall = nil;
}

static char *
ioctl2str(int cmd)
{
	int i;
	static char buf[32];

	for(i = 0; Ioctlnames[i].s; i++)
		if(Ioctlnames[i].n == cmd)
			return Ioctlnames[i].s;
	snprint(buf, sizeof(buf), "0x%x", cmd);
	return buf;
}

static void
dotrace(Linuxcall *c, Ureg *u)
{
	Fmt f;
	int n;
	char *p, *str;
	ulong a[] = { u->bx, u->cx, u->dx, u->si, u->di, u->bp };

	if(strcmp(c->args, "-") == 0){
		trace("%s()", c->name);
		return;
	}

	if(strcmp(c->args, "?") == 0){
		trace("%s(%ux, %ux, %ux, %ux, %ux, %ux)",
			c->name, u->bx, u->cx, u->dx, u->si, u->di, u->bp);
		return;
	}

	if(c == nil || c->name == nil || c->args == nil || c->func == nil)
		sysfatal("internal error: bad syscall %lud\n", u->ax);

	fmtstrinit(&f);
	fmtprint(&f, "%s(", c->name);
	for(n = 0, p = c->args; *p; p++, n++){
		if(p != c->args)
			fmtprint(&f, ", ");

		switch(*p){
		case 'F':	// file descriptor or -1
		case 'f':	// file descriptor
			fmtprint(&f, "fd=%ld", a[n]);
			break;
		case 'b':	// random buffer
		case 'a':	// argv or envp like array
		case 'B':	// random buffer or nil
			fmtprint(&f, "buf=0x%lux", a[n]);
			break;
		case 'l':	// buffer size
			fmtprint(&f, "nbuf=%lud", a[n]);
			break;
		case 'o':	// open mode
			fmtprint(&f, "omode=%lud", a[n]);
			break;
		case 'm':	// file permissions
			fmtprint(&f, "perm=0%luo", a[n]);
			break;
		case 'p':	// existing file path
			fmtprint(&f, "path=%s", (char *)a[n]);
			break;
		case 'P':	// new file path
			fmtprint(&f, "newpath=%s", (char *)a[n]);
			break;
		case 'N':	// existing file path or nil
			if(a[n])
				fmtprint(&f, "path=%s", (char *)a[n]);
			else
				fmtprint(&f, "path=<nil>");
			break;
		case 'u':	// generic unsigned integer
			fmtprint(&f, "n=%lud", a[n]);
			break;
		case 'd':	// generic signed integer
			fmtprint(&f, "n=%lud", a[n]);
			break;
		case 'x':	// generic hex integer
			fmtprint(&f, "n=0x%lux", a[n]);
			break;
		case 'w':	// seek whence
			fmtprint(&f, "whence=%lud", a[n]);
			break;
		case 's':	// seek offset
			fmtprint(&f, "off=%lud", a[n]);
			break;
		case 'U':	// uid
			fmtprint(&f, "uid=%ld", a[n]);
			break;
		case 'G':	// gid
			fmtprint(&f, "uid=%ld", a[n]);
			break;
		case 'I':	// pid
			fmtprint(&f, "pid=%ld", a[n]);
			break;
		case 'T':	// tid
			fmtprint(&f, "tid=%ld", a[n]);
			break;
		case 'S':	// signal
			fmtprint(&f, "sig=%S", (int)a[n]);
			break;
		case 't':	// timeout
			fmtprint(&f, "timeout=%ld", a[n]);
			break;
		case 'i':	// ioctl command number
			fmtprint(&f, "cmd=%s", ioctl2str(a[n]));
			break;
		default:
			fprint(2, "dotrace: '%c' bad arg fmt char \n", *p);
			if(debug)
				abort();
			break;
		}
	}
	fmtprint(&f, ")");
	str = fmtstrflush(&f);
	trace("%s", str);
	free(str);
}

void
linuxcall(void)
{
	Uproc *p;
	Ureg *u;
	Linuxcall *c;

	p = current;
	u = p->ureg;

	if(u->ax > nelem(linuxcalltab)){
		kprint("%d - unsupported syscall ID\n", u->ax);
		linuxret(-ENOSYS);
		return;
	}

	c = &linuxcalltab[u->ax];
	if(p->restart->syscall)
		trace("linuxcall(): restarting %s",c->name);
	else
		dotrace(c, u);

	p->sysret = linuxret;
	p->syscall = c->name;
	p->calltab = c;

	c->calls++;
	cycles(&p->enter);
	if(_tos->pcycles)
		p->pcycles = _tos->pcycles;
	if(_tos->kcycles)
		p->kcycles = _tos->kcycles;

	linuxret(trampoline(u, c->func));
}

int
_kernel_vsyscall(uint idx, uint a1, uint a2, uint a3, uint a4, uint a5, uint a6)
{
	Ureg ureg, *u;
	Uproc *p;

	p = current;
	memset(&ureg, 0, sizeof(ureg));
	u = p->ureg = &ureg;
	u->ax = idx;
	u->bx = a1;
	u->cx = a2;
	u->dx = a3;
	u->si = a4;
	u->di = a5;
	u->bp = a6;
	linuxcall();
	return u->ax;

}

char *
syscallstats(void)
{
	Fmt f;
	char buf[128];
	vlong mhz;
	Linuxcall *c;

	mhz = _tos->cyclefreq / 1000000LL;

	fmtstrinit(&f);
	fmtprint(&f, "%-24s %11s %22s %22s %22s %22s    (emu+plan9 μs)\n",
		"#name", "calls", "min", "max", "total", "mean");

	for(c = linuxcalltab; c < &linuxcalltab[nelem(linuxcalltab)]; c++){
		if(c->calls == 0)
			continue;

		fmtprint(&f, "%-24s %11lld ", c->name, c->calls);
		snprint(buf, sizeof(buf), "%lld+%lld+%lld",
			c->min[User]/mhz, (c->min[Real] - c->min[Kern])/mhz, c->min[Kern]/mhz);
		fmtprint(&f, "%22s ", buf);

		snprint(buf, sizeof(buf), "%lld+%lld+%lld",
			c->max[User]/mhz, (c->max[Real] - c->max[Kern])/mhz, c->max[Kern]/mhz);
		fmtprint(&f, "%22s ", buf);

		snprint(buf, sizeof(buf), "%lld+%lld+%lld",
			c->tot[User]/mhz, (c->tot[Real] - c->tot[Kern])/mhz, c->tot[Kern]/mhz);
		fmtprint(&f, "%22s ", buf);

		snprint(buf, sizeof(buf), "%lld+%lld+%lld",
			(c->tot[User]/mhz)/c->calls,
			((c->tot[Real] - c->tot[Kern])/mhz)/c->calls, (c->tot[Kern]/mhz)/c->calls);
		fmtprint(&f, "%22s\n", buf);

	}
	return fmtstrflush(&f);
};

