#include <u.h>
#include <libc.h>
#include <tos.h>
#include <ureg.h>
#include <pool.h>
#include <ip.h>
#include "dat.h"
#include "fns.h"
#include "linux.h"


static void
die(void)
{
	exits(nil);
}

static char *
xlatezone(char *buf)
{
	int n, stddiff, dstdiff;
	char *s, *a[16], *std, *dst;
	int first, last;


	if((n = tokenize(buf, a, nelem(a))) < 4)
		return buf;

	std = a[0];
	stddiff = atoi(a[1]) / 3600;
	dst = a[2];
	dstdiff = -atoi(a[3]) / 3600;

	if(n < 6){
		s = smprint("%s%+d%s%+d", std, stddiff, dst, dstdiff);
		free(buf);
		return s;
	}

	first = localtime(atol(a[4]))->yday;
	last = localtime(atol(a[5]))->yday;
	s = smprint("%s%+d%s%+d,%d,%d", std, stddiff, dst, dstdiff, first, last);
	free(buf);
	return s;

}

	
static char**
readenv(void)
{
	char **env;
	char *envdir;
	int fd, n, i, c;
	Dir *d;

	envdir = "/env/";
	if((fd = open(envdir, OREAD)) < 0){
		kprint("%s cannot open dir\n", envdir);
		return nil;
	}
	n = dirreadall(fd, &d);
	close(fd);
	env = kmalloc(sizeof(env[0]) * (n + 1));
	c = 0;
	for(i=0; i<n; i++){
		char *v;
		char *k;

		k = d[i].name;

		// filter out some stuff...
		if(strncmp(k, "fn#", 3) == 0)
			continue;
			
		if(strcmp(k, "0")==0)
			continue;

		if((v = getenv(d[i].name)) == nil)
			continue;
		if(strcmp(k, "timezone") == 0){
			v = xlatezone(v);
			k = "TZ";
		}
		if((env[c] = ksmprint("%s=%s", k, v)) == nil)
			continue;
		free(v);

		c++;
	}
	env[c] = 0;

	free(d);
	return env;
}

struct onstackargs
{
	long		*stk;
	void		*arg;
	int		(*func)(void *);
	int		ret;
	jmp_buf	jmp;
};

int
onstack(long *stk, int (*func)(void *), void *arg)
{
	struct onstackargs a, *args;
	jmp_buf jmp;
	long *sp;

	sp = (long*)&a;
	if((long*)sp >= stk && (long*)sp < stk+(KSTACK / sizeof(long)))
		return func(arg);

	if(args = (struct onstackargs*)setjmp(jmp)){
		args->ret = onstack(args->stk, args->func, args->arg);
		longjmp(args->jmp, 1);
	}

	sp = &stk[(KSTACK / sizeof(long))-16];
	jmp[JMPBUFSP] = (long)sp;

	memset(stk, 0, KSTACK);

	args = &a;
	args->stk = stk;
	args->func = func;
	args->arg = arg;

	if(!setjmp(args->jmp))
		longjmp(jmp, (int)args);

	return args->ret;
}

#pragma profile off

static void
proff(void (*fn)(void*), void *arg)
{
	if(_tos->prof.what == 0){
		fn(arg);
	}else{
		prof(fn, arg, 2000, _tos->prof.what);
	}
}

static void
profexitjmpfn(void *arg)
{
	/*
	 * we are now called by the profiling function on the profstack.
	 * save the current continuation so we can return here on exit.
	 */
	if(!setjmp(exitjmp))
		longjmp((long*)arg, 1); /* return from profme() */
}

static int
profmeprofstack(void *arg)
{
	proff(profexitjmpfn, arg);
	for(;;) die();
}

#pragma profile on

static long *profstack;

void
profme(void)
{
	jmp_buf j;

	if(!setjmp(j))
		onstack(profstack, profmeprofstack, j);
}


static void
vpanic(char *msg, va_list arg)
{
	char buf[32];
	int fd;

	fprint(2, "PANIC: ");
	vfprint(2, msg, arg);
	fprint(2, "\n");

	if(debug)
		abort();

	snprint(buf, sizeof(buf), "/proc/%d/notepg", getpid());
	if((fd = open(buf, OWRITE)) >= 0){
		write(fd, "kill", 4);
		close(fd);
	}
	exits("panic");
}

int
setlabel(char **argv)
{
	int n, fd;
	char **a, *s;

	n = 0;
	for(a = argv; *a; a++)
		n += strlen(*a) +1;

	s = mallocz(n, 1);
	if(s == nil)
		return -1;

	for(a = argv; *a; a++){
		strcat(s, *a);
		if(a[1])
			strcat(s, " ");
	}

	fd = open("/dev/label", OWRITE);
	if(fd == -1)
		return -1;
	fprint(fd, "%s", s);
	close(fd);
	free(s);
	return 0;
}

void
panic(char *msg, ...)
{
	va_list arg;

	va_start(arg, msg);
	vpanic(msg, arg);
	va_end(arg);
}

static void
usage(void)
{
	fprint(2, "usage: linuxemu [-d] [-u uid] [-g gid] cmd [args]\n");
	exits("usage");
}

struct mainstack
{
	long		profstack[KSTACK / sizeof(long)];
	long		kstack[KSTACK / sizeof(long)];
	Uproc	*proc;
	jmp_buf	exitjmp;
};

void
main(int argc, char *argv[])
{
	struct mainstack ms;
	char *root;
	int err, uid, gid, fd;

	fmtinstall('E', Efmt);
	fmtinstall('S', Sfmt);
	fmtinstall('V', eipfmt);

	uid = 0;
	gid = 0;
	debug = 0;

	root = "/";
	ARGBEGIN {
	case 'r':
		root = EARGF(usage());
		break;
	case 'd':
		debug++;
		break;
	case 'm':
		mainmem->flags |= POOL_PARANOIA|POOL_ANTAGONISM|POOL_NOREUSE;
		break;
	case 'u':
		uid = atoi(EARGF(usage()));
		break;
	case 'g':
		gid = atoi(EARGF(usage()));
		break;
	default:
		usage();
	} ARGEND

	if(argc < 1)
		usage();

	rootdevinit(root);
	devdirinit();
	procdevinit(root);
	ptydevinit();
	consdevinit();
	dspdevinit();
	miscdevinit();
	sockdevinit();
	pipedevinit();
	etcdevinit();

	kstack = ms.kstack;
	profstack = ms.profstack;
	exitjmp = ms.exitjmp;
	pcurrent = &ms.proc;
	current = nil;

	if(setjmp(exitjmp))
		die();

	initproc(root);
	current->uid = uid;
	current->gid = gid;

	/* emulated console */
	if((fd = sys_open("/dev/console", O_RDWR, 0)) != 0)
		fprint(2, "cant open /dev/console (stdin) - %E\n", fd);
	if((err = sys_dup(fd)) != 1)
		fprint(2, "cant dup (stdout) - %E\n", err);
	if((err = sys_dup(fd)) != 2)
		fprint(2, "cant dup (stderr) - %E\n", err);

	sys_fcntl(0, F_SETFD, 0);
	sys_fcntl(1, F_SETFD, 0);
	sys_fcntl(2, F_SETFD, 0);

	if(isaconsole(0))
		setlabel(argv);

	err = sys_execve(*argv, argv, readenv());

	fprint(2, "%s - %E\n", *argv, err);
	longjmp(exitjmp, 1);
}
