#include <u.h>
#include <libc.h>
#include <ureg.h>
#include <mp.h>
#include <libsec.h>

#include "dat.h"
#include "fns.h"
#include "linux.h"

typedef struct Miscfile Miscfile;
struct Miscfile
{
	Ufile;
	int	q;
	vlong lastoff;
	char	*data;
	int	ndata;
	Framebuffer *fb;
};

enum
{
	Qnull,
	Qzero,
	Qfull,
	Qrandom,
	Qurandom,
	Qinitctl,

	Qstdin,
	Qstdout,
	Qstderr,
	Qfd,
	Qtty,
	Qconsctl,
//	Qmem,
	Qfips,
	Qfb0,
};

static struct {
	int mode;
	char *name;
	int rdev;
	char *link;
} miscdevtab[] = {
	[Qnull]		0666|S_IFCHR,	"null",		makedev(1, 3),	"",
	[Qzero]		0666|S_IFCHR,	"zero",		makedev(1, 5),	"",
	[Qfull]		0666|S_IFCHR,	"full",		makedev(1, 7),	"",
	[Qrandom]	0666|S_IFCHR,	"random",	makedev(1, 8),	"",
	[Qurandom]	0666|S_IFCHR,	"urandom",	makedev(1, 9),	"",
	[Qinitctl]	0755|S_IFIFO,	"initctl",	makedev(1, 99), "",

	[Qstdin]	0644|S_IFLNK,	"stdin",	0, 		"/proc/self/fd/0",
	[Qstdout]	0644|S_IFLNK,	"stdout",	0, 		"/proc/self/fd/1",
	[Qstderr]	0644|S_IFLNK,	"stderr",	0, 		"/proc/self/fd/2",
	[Qfd]		0755|S_IFLNK,	"fd",		0, 		"/proc/self/fd",
	[Qtty]		0755|S_IFLNK,	"tty",		0, 		"/proc/self/fd/0",
	[Qconsctl]	0755|S_IFLNK,	"consctl",	0, 		"/n/wsys/consctl",
// no /proc/self/mem yet
// 	[Qmem]		0755|S_IFLNK,	"mem",		0, 		"/proc/self/mem",
	[Qfips]		0644|S_IFREG,	"fips_enabled",	0,		"0",
	[Qfb0]		0666|S_IFCHR,	"fb0",		makedev(29, 0),	"",
};


static int
path2q(char *path)
{
	int q;
	char *p;

	if((p = strrchr(path, '/')) == nil)
		return -1;
	for(q = 0; q < nelem(miscdevtab); q++)
		if(strcmp(p+1, miscdevtab[q].name) == 0)
			return q;
	return -1;
}

static int
openmisc(char *path, int omode, int, Ufile **pf)
{
	void *fbmem;
	ulong fbsiz;
	Miscfile *mf;
	Framebuffer *fb;
	int q;

	if((q = path2q(path)) < 0)
		return -ENOENT;
	if((miscdevtab[q].mode & S_IFLNK) == S_IFLNK)
		return -SYMLINK;

	fb = nil;
	fbsiz = 0;
	fbmem = nil;
	if(q == Qfb0){
		if((fb = fb_open(&fbmem, &fbsiz)) == nil)
			return -EIO;
	}

	mf = kmallocz(sizeof(*mf), 1);
	mf->ref = 1;
	mf->omode = omode;
	mf->path = kstrdup(path);
	mf->fd = -1;
	mf->dev = MISCDEV;
	mf->q = q;
	mf->fb = fb;
	mf->fbmem = fbmem;
	mf->fbsiz = fbsiz;

	*pf = mf;
	return 0;
}

static int
closemisc(Ufile *file)
{
	Miscfile *mf;

	mf = (Miscfile *)file;
	if(mf->q == Qfb0)
		fb_close(mf->fb);
	return 0;
}

static int
readstr(ulong off, char *buf, ulong n, char *str)
{
	int size;

	size = strlen(str);
	if(off >= size)
		return 0;
	if(off+n > size)
		n = size-off;
	memmove(buf, str+off, n);
	return n;
}


static int
readmisc(Ufile *file, void *buf, int len, vlong off)
{
	Miscfile *mf = (Miscfile *)file;

	switch(mf->q){
	case Qnull:
		return 0;
	case Qzero:
	case Qfull:
		memset(buf, 0, len);
		return len;
	case Qrandom:
		genrandom(buf, len);
		return len;
	case Qurandom:
		prng(buf, len);
		return len;
	case Qfips:
		return readstr(off, buf, len, "0\n");
	case Qfb0:
		return fb_read(mf->fb, buf, len, off);
	default:
		return -EIO;
	}
}

static int
writemisc(Ufile *file, void *buf, int len, vlong off)
{
	Miscfile *mf = (Miscfile *)file;

	switch(mf->q){
	case Qnull:
	case Qzero:
	case Qrandom:
	case Qurandom:
		return len;
	case Qfull:
		return -ENOSPC;
	case Qfb0:
		return fb_write(mf->fb, buf, len, off);
	default:
		return -EIO;
	}
}

static int
readdirmisc(Ufile *, Udirent **pd)
{
	int q, n;

	n = 0;
	for(q = 0; q < nelem(miscdevtab); q++){
		if((*pd = newdirent(miscdevtab[q].name, miscdevtab[q].mode, mkino(MISCDEV, q))) == nil)
			break;
		pd = &((*pd)->next);
		n++;
	}
	return n;
}

static int
statmisc(char *path, int link, Ustat *s)
{
	int q;

	if((q = path2q(path)) < 0)
		return -ENOENT;
	if(! link && (miscdevtab[q].mode & S_IFLNK) == S_IFLNK)
		return -SYMLINK;
	s->mode = miscdevtab[q].mode;
	if((s->mode & S_IFLNK) == S_IFLNK)
		s->size = strlen(miscdevtab[q].link);
	s->rdev = miscdevtab[q].rdev;
	s->ino = mkino(MISCDEV, q);
	s->dev = MISCDEV;
	s->atime = s->mtime = s->ctime = boottime/1000000000LL;
	return 0;
}

static vlong
sizemisc(Ufile *file)
{
	Miscfile *mf = (Miscfile *)file;

	if((miscdevtab[mf->q].mode & S_IFLNK) == S_IFLNK)
		return strlen(miscdevtab[mf->q].link);
	return 0;
}

static int
fstatmisc(Ufile *file, Ustat *s)
{
	return statmisc(file->path, 0, s);
};


static int
readlinkmisc(char *path, char *buf, int len)
{
	int n, q;

	if((q = path2q(path)) < 0)
		return -ENOENT;
	if((miscdevtab[q].mode & S_IFLNK) != S_IFLNK)
		return -EINVAL;
	n = strlen(miscdevtab[q].link);
	if(len > n)
		len = n;
	memcpy(buf, miscdevtab[q].link, len);
	return len;
}

static int
ioctlmisc(Ufile *file, int cmd, void *arg)
{
	Miscfile *mf;
	enum {
		FBIOGET_VSCREENINFO = 0x4600,
		FBIOPUT_VSCREENINFO = 0x4601,
		FBIOGET_FSCREENINFO = 0x4602,
		FBIOGETCMAP = 0x4604,
		FBIOPUTCMAP = 0x4605,
		FBIOPAN_DISPLAY = 0x4606,
		FBIOGET_MONITORSPEC = 0x460C,
		FBIOPUT_MONITORSPEC = 0x460D,
		FBIOSWITCH_MONIBIT = 0x460E,
		FBIOGET_CON2FBMAP = 0x460F,
		FBIOPUT_CON2FBMAP = 0x4610,
		FBIOBLANK = 0x4611,
		FBIO_ALLOC = 0x4613,
		FBIO_FREE = 0x4614,
		FBIOGET_GLYPH = 0x4615,
		FBIOGET_HWCINFO = 0x4616,
		FBIOPUT_MODEINFO = 0x4617,
		FBIOGET_DISPINFO = 0x4618,
	};

	mf = (Miscfile *)file;
	if(mf->q != Qfb0)
		return -ENOENT;

	switch(cmd){
	case FBIOPUT_VSCREENINFO:
		return fb_putvarinfo(mf->fb, arg);
	case FBIOGET_VSCREENINFO:
		return fb_getvarinfo(mf->fb, arg);
	case FBIOGET_FSCREENINFO:
		return fb_getfixinfo(mf->fb, arg);
	default:
		print("ioctl: 0x%x %p\n", cmd, arg);
		break;
	}
	return -ENOENT;
}

static Udev miscdev =
{
	.name = "misc",
	.open = openmisc,
	.read = readmisc,
	.size = sizemisc,
	.write = writemisc,
	.close = closemisc,
	.stat = statmisc,
	.fstat = fstatmisc,
	.readdir = readdirmisc,
	.readlink = readlinkmisc,
	.ioctl = ioctlmisc,
};

void miscdevinit()
{
	devtab[MISCDEV] = &miscdev;

	fsmount(&miscdev, "/dev/null");
	fsmount(&miscdev, "/dev/zero");
	fsmount(&miscdev, "/dev/full");
	fsmount(&miscdev, "/dev/random");
	fsmount(&miscdev, "/dev/urandom");
	fsmount(&miscdev, "/dev/initctl");

	fsmount(&miscdev, "/dev/stdin");
	fsmount(&miscdev, "/dev/stdout");
	fsmount(&miscdev, "/dev/stderr");
	fsmount(&miscdev, "/dev/fd");
	fsmount(&miscdev, "/dev/consctl");
	fsmount(&miscdev, "/dev/tty");
//	fsmount(&miscdev, "/dev/mem");
	fsmount(&miscdev, "/proc/sys/crypto/fips_enabled");	 /* FIXME: bodge */
	fsmount(&miscdev, "/dev/fb0");

	srand(truerand());
}
