#include <u.h>
#include <libc.h>
#include <ureg.h>
#include <ip.h>
#include "dat.h"
#include "fns.h"
#include "linux.h"
#include "socket.h"

enum {	/* netlink message flags */
	NLM_F_REQUEST	= 0x001,	/* It is request message */
	NLM_F_MULTI	= 0x002,	/* Multipart message, terminated by NLMSG_DONE */
	NLM_F_ACK	= 0x004,	/* Reply with ack, with zero or error code */
	NLM_F_ECHO	= 0x008,	/* Echo this request */

					/* GET requests */
	NLM_F_ROOT	= 0x100,	/* specify tree	root */
	NLM_F_ATOMIC	= 0x400,	/* atomic GET */
	NLM_F_DUMP	= 0x300,

					/* NEW requests */
	NLM_F_REPLACE	= 0x100,	/* Override existing */
	NLM_F_EXCL	= 0x200,	/* Do not touch, if it exists */
	NLM_F_CREATE	= 0x400,	/* Create, if it does not exist	*/
	NLM_F_APPEND	= 0x800,	/* Add to end of list */
};

struct nlmsghdr {			/* netlink message header */
	uint nlmsg_len;
	ushort nlmsg_type;
	ushort nlmsg_flags;
	uint nlmsg_seq;
	uint nlmsg_pid;
};

struct rtattr {
	ushort rta_len;
	ushort rta_type;
};

struct ifinfomsg
{
	uchar	ifi_family;
	uchar	__ifi_pad;
	ushort	ifi_type;		/* ARPHRD_* */
	int	ifi_index;		/* Link index	*/
	uint	ifi_flags;		/* IFF_* flags	*/
	uint	ifi_change;		/* IFF_* change mask */
};

#define NLMSG_ALIGNTO	4
#define NLMSG_ALIGN(len) ( ((len)+NLMSG_ALIGNTO-1) & ~(NLMSG_ALIGNTO-1) )
#define NLMSG_HDRLEN	 ((int) NLMSG_ALIGN(sizeof(struct nlmsghdr)))
#define NLMSG_LENGTH(len) ((len)+NLMSG_ALIGN(NLMSG_HDRLEN))
#define NLMSG_SPACE(len) NLMSG_ALIGN(NLMSG_LENGTH(len))
#define NLMSG_DATA(nlh)  ((void*)(((char*)nlh) + NLMSG_LENGTH(0)))
#define NLMSG_NEXT(nlh,len)	 ((len) -= NLMSG_ALIGN((nlh)->nlmsg_len), \
				  (struct nlmsghdr*)(((char*)(nlh)) + NLMSG_ALIGN((nlh)->nlmsg_len)))
#define NLMSG_OK(nlh,len) ((len) >= (int)sizeof(struct nlmsghdr) && \
			   (nlh)->nlmsg_len >= sizeof(struct nlmsghdr) && \
			   (nlh)->nlmsg_len <= (len))
#define NLMSG_PAYLOAD(nlh,len) ((nlh)->nlmsg_len - NLMSG_SPACE((len)))

#define NLMSG_NOOP		0x1	/* Nothing.		*/
#define NLMSG_ERROR		0x2	/* Error		*/
#define NLMSG_DONE		0x3	/* End of a dump	*/
#define NLMSG_OVERRUN		0x4	/* Data lost		*/

#define NLMSG_MIN_TYPE		0x10	/* < 0x10: reserved control messages */
/* Macros to handle rtattributes */

#define RTA_ALIGNTO	4
#define RTA_ALIGN(len) ( ((len)+RTA_ALIGNTO-1) & ~(RTA_ALIGNTO-1) )
#define RTA_OK(rta,len) ((len) >= (int)sizeof(struct rtattr) && \
			 (rta)->rta_len >= sizeof(struct rtattr) && \
			 (rta)->rta_len <= (len))
#define RTA_NEXT(rta,attrlen)	((attrlen) -= RTA_ALIGN((rta)->rta_len), \
				 (struct rtattr*)(((char*)(rta)) + RTA_ALIGN((rta)->rta_len)))
#define RTA_LENGTH(len)	(RTA_ALIGN(sizeof(struct rtattr)) + (len))
#define RTA_SPACE(len)	RTA_ALIGN(RTA_LENGTH(len))
#define RTA_DATA(rta)   ((void*)(((char*)(rta)) + RTA_LENGTH(0)))
#define RTA_PAYLOAD(rta) ((int)((rta)->rta_len) - RTA_LENGTH(0))

enum
{
	IFLA_UNSPEC,
	IFLA_ADDRESS,
	IFLA_BROADCAST,
	IFLA_IFNAME,
	IFLA_MTU,
	IFLA_LINK,
	IFLA_QDISC,
	IFLA_STATS,
	IFLA_COST,
	IFLA_PRIORITY,
	IFLA_MASTER,
	IFLA_WIRELESS,		/* Wireless Extension event - see wireless.h */
	IFLA_PROTINFO,		/* Protocol specific information for a link */
	IFLA_TXQLEN,
	IFLA_MAP,
	IFLA_WEIGHT,
	IFLA_OPERSTATE,
	IFLA_LINKMODE,
	__IFLA_MAX
};

#define IFLA_MAX (__IFLA_MAX - 1)

#define IFLA_RTA(r)  ((struct rtattr*)(((char*)(r)) + NLMSG_ALIGN(sizeof(struct ifinfomsg))))
#define IFLA_PAYLOAD(n) NLMSG_PAYLOAD(n,sizeof(struct ifinfomsg))

static char *typenames[] = {
	[16]	"new link",
	[17]	"del link",
	[18]	"get link",
	[19]	"set link",

	[20]	"new addr",
	[21]	"del addr",
	[22]	"get addr",

	[24]	"new route",
	[25]	"del route",
	[26]	"get route",

	[28]	"new neigh",
	[29]	"del neigh",
	[30]	"get neigh",

	[32]	"new rule",
	[33]	"del rule",
	[34]	"get rule",

	[36]	"new q disc",
	[37]	"del q disc",
	[38]	"get q disc",

	[40]	"new t class",
	[41]	"del t class",
	[42]	"get t class",

	[44]	"new t filter",
	[45]	"del t filter",
	[46]	"get t filter",

	[48]	"new action",
	[49]	"del action",
	[50]	"get action",

	[52]	"new prefix",
	[54]	"get prefix",

	[58]	"get multicast",

	[62]	"get anycast",

	[64]	"new neigh tbl",
	[66]	"get neigh tbl",
	[67]	"set neigh tbl",

	[72]	"new addr label",
	[73]	"del addr label",
	[74]	"get addr label",

};
static void print_char(struct rtattr *attr, const char *name)
{
	kprint("  attr %s (len %jd) %s\n", name, RTA_PAYLOAD(attr),
						(char *) RTA_DATA(attr));
}

static void print_byte(struct rtattr *attr, const char *name)
{
	kprint("  attr %s (len %jd) 0x%02x\n", name, RTA_PAYLOAD(attr),
					*((unsigned char *) RTA_DATA(attr)));
}

static void print_attr(struct rtattr *attr, const char *name)
{
	if (name)
		kprint("  attr %s (len %jd)\n", name, RTA_PAYLOAD(attr));
	else
		kprint("  attr %d (len %jd)\n",
					attr->rta_type, RTA_PAYLOAD(attr));
}


static void rtnl_link(struct nlmsghdr *hdr)
{
	struct ifinfomsg *msg;
	struct rtattr *attr;
	int bytes;

	msg = (struct ifinfomsg *) NLMSG_DATA(hdr);
	bytes = IFLA_PAYLOAD(hdr);

	kprint("ifi_index %d ifi_flags 0x%04x", msg->ifi_index, msg->ifi_flags);

	for (attr = IFLA_RTA(msg); RTA_OK(attr, bytes);
					attr = RTA_NEXT(attr, bytes)) {
		switch (attr->rta_type) {
		case IFLA_ADDRESS:
			print_attr(attr, "address");
			break;
		case IFLA_BROADCAST:
			print_attr(attr, "broadcast");
			break;
		case IFLA_IFNAME:
			print_char(attr, "ifname");
			break;
		case IFLA_MTU:
			print_attr(attr, "mtu");
			break;
		case IFLA_LINK:
			print_attr(attr, "link");
			break;
		case IFLA_QDISC:
			print_attr(attr, "qdisc");
			break;
		case IFLA_STATS:
			print_attr(attr, "stats");
			break;
		case IFLA_COST:
			print_attr(attr, "cost");
			break;
		case IFLA_PRIORITY:
			print_attr(attr, "priority");
			break;
		case IFLA_MASTER:
			print_attr(attr, "master");
			break;
		case IFLA_WIRELESS:
			print_attr(attr, "wireless");
			break;
		case IFLA_PROTINFO:
			print_attr(attr, "protinfo");
			break;
		case IFLA_TXQLEN:
			print_attr(attr, "txqlen");
			break;
		case IFLA_MAP:
			print_attr(attr, "map");
			break;
		case IFLA_WEIGHT:
			print_attr(attr, "weight");
			break;
		case IFLA_OPERSTATE:
			print_byte(attr, "operstate");
			break;
		case IFLA_LINKMODE:
			print_byte(attr, "linkmode");
			break;
		default:
			print_attr(attr, nil);
			break;
		}
	}
}

int
nl_write(Socket *sock, void *buf, int len, vlong off)
{
	char *flags, *type;
	struct nlmsghdr	*h;
	struct rtattr *r;

	fprint(2, "nl_write sock=%p buf=%p len=%d off=%lld\n", sock, buf, len, off);
	if(len >= sizeof(struct nlmsghdr)){
		h = buf;
rtnl_link(h);
kprint("\n\n");
		switch(h->nlmsg_flags & 0x0f){
		case NLM_F_REQUEST:	flags = "req";		break;
		case NLM_F_MULTI:	flags = "multi";	break;
		case NLM_F_ACK:		flags = "ack";		break;
		case NLM_F_ECHO:	flags = "echo";		break;
		default:		flags = "?";		break;
		}

		type = "?";
		if(h->nlmsg_type < nelem(typenames) && typenames[h->nlmsg_type])
			type = typenames[h->nlmsg_type];

		fprint(2, "  nlmsghdr: sz=%d len=%d type=%d=%s flags=%x=%s seq=%d pid=%d\n",
			sizeof(struct nlmsghdr), h->nlmsg_len, h->nlmsg_type, type,
			h->nlmsg_flags, flags, h->nlmsg_seq, h->nlmsg_pid);

		len -= sizeof(struct nlmsghdr);
		buf = &h[1];
	}

	if(len >= sizeof(struct rtattr)){
		r = buf;
		fprint(2, "    rtattr: sz=%d len=%d type=%d\n",
			sizeof(struct rtattr), r->rta_len, r->rta_type); 
	}
	return len;
}

int
nl_read(Socket *sock, void *buf, int len, vlong off)
{
	fprint(2, "nl_read sock=%p buf=%p len=%d off=%lld\n", sock, buf, len, off);

{
static int x = 0;
if(x++ > 10)
	abort();
}

	return -ENOSYS;		
}

