#include <u.h>
#include <libc.h>
#include <ureg.h>
#include "dat.h"
#include "fns.h"
#include "linux.h"

typedef struct Cache Cache;
struct Cache {
	Cache *next;
	char *path;	
	char *dest;
	int err;
	vlong hits;
	long atime;	
};

static struct {
	RWLock;
	Cache *root;
} cachealloc;

void
updatecache(int err, char *path, char *dest)
{
	Cache *c;

	if(err == 0 && dest == nil)
		return;		/* no advantage in caching this */

	wlock(&cachealloc);
	for(c = cachealloc.root; c; c = c->next)
		if(strcmp(c->path, path) == 0)
			break;

	/* cache entry unchanged */
	if(c && c->err == err){
		if(dest == nil && c->dest == nil){
			c->atime = time(nil);
			wunlock(&cachealloc);
			return;
		}
		if(dest && c->dest && strcmp(dest, c->dest) == 0){
			c->atime = time(nil);
			wunlock(&cachealloc);
			return;
		}
	}


	/* add entry */
	if(c == nil){
		c = kmallocz(sizeof(Cache)+strlen(path)+1, 1);
		c->path = (char *)&c[1];
		strcpy(c->path, path);
		if(dest)
			c->dest = kstrdup(dest);
		c->atime = time(nil);

		c->next = cachealloc.root;
		c->err = err;
		c->hits = 1;
		cachealloc.root = c;
		wunlock(&cachealloc);
		if(debug)
			trace("	updcache: %s %E %s new", path, err, dest);
		return;
	}


	/* modify entry */
	if(debug)
		trace("	updcache: %s %E->%E %s->%s old", path, c->err, err, c->dest, dest);
	c->err = err;
	c->atime = time(nil);
	free(c->dest);
	if(dest)
		c->dest = kstrdup(dest);
	else
		c->dest = nil;
	wunlock(&cachealloc);
} 
 
int
lookcache(char **path, int check)
{
	int err;
	Cache *c;

	if(check != Pfull)		/* FIXME:  bodge for now */
		return 0;

	err = 0;
	rlock(&cachealloc);
	for(c = cachealloc.root; c; c = c->next)
		if(strcmp(c->path, *path) == 0)
			break;

	if(c){
		c->hits++;
		c->atime = time(0);
		switch(c->err){
		case 0:			/* is a symlink */
			if(debug)
				trace("	lookcache %s -> %s", *path, c->dest);
			if(c->dest){
				free(*path);
				*path = kstrdup(c->dest);
			}
			break;
		case -ESRCH:		/* process doesn't exist */
		case -ENOENT:		/* file doesn't exist */
			err = c->err;
			if(debug)
				trace("	cachelook %s <not found>", *path);
			break;
		default:
			print("lookcache: err=%E path=%s dest=%s unknown err\n",
				c->err, c->path, c->dest);
			break;
		}
	}
	runlock(&cachealloc);
	return err;
}

int
flushcache(char *path)
{
	Cache *c, **p;

	wlock(&cachealloc);
	p = &cachealloc.root;
	for(c = cachealloc.root; c; c = c->next){
		if(strcmp(path, c->path) == 0){
			*p = c->next;
			break;
		}
		p = &c->next;
	}

	wunlock(&cachealloc);

	if(c == nil)
		return -1;
	free(c->dest);
	free(c);
	if(debug)
		trace("	flushcache: %s", path);
	return 0;
}

char *
cacheentries(void)
{
	long now;
	Fmt f;
	Cache *c;

	now = time(nil);
	fmtstrinit(&f);
	fmtprint(&f, "%-8s %-6s %-6s %-32s %s\n", "#type", "hits", "age", "path", "symlink");

	rlock(&cachealloc);
	for(c = cachealloc.root; c; c = c->next)
		if(c->dest)
			fmtprint(&f, "%8E %-6lld %-6ld %-32s %s\n",	
				c->err, c->hits, now - c->atime, c->path, c->dest);
		else
			fmtprint(&f, "%8E %-6lld %-6ld %s\n",	
				c->err, c->hits, now - c->atime, c->path);
	runlock(&cachealloc);

	return fmtstrflush(&f);
}

			
/*
 * Note: no optimisation here, we rescan the path after crossing every symlink,
 * strictly we can rescan from "here" each time if the substution is relative
 * but we rely on the file metadata cache and make this code simpler.
 */
enum {
	Maxlink = 8192
};

static int
deref(char **buf, char *here, char *next)
{
	int n;
	char *npath, *slash, *new;

	/* FIXME: might be nice to add genbuf to the proc structure, we could use this here */

	new = kmalloc(Maxlink);
	if((n = fsreadlink(*buf, new, Maxlink-1)) < 0){
		free(new);
		return n;
	}
	new[n] = 0;

	slash = (*next)? "/": "";
	if(*new == '/')		/* absolute path - new prefix */
		npath = ksmprint("%s%s%s", new, slash, next);
	else			/* relative path - insert into buffer */
		npath = ksmprint("%.*s/%s%s%s", (int)(here-*buf), *buf, new, slash, next);

	free(new);
	free(*buf);
	*buf = npath;
	return 0;
}

/*
 * we take care of the two filesystems which have virtual symlinks 
 * (procdev and miscdev) and don't cache symlinks which redirect
 * through them.
 */
int
retry(int err, char **path, int check)
{
	Ustat ps;
	char *p, *here, *buf;
	int e, dynlinks, justlinked, redir;

	trace("	retry err=%E path=%s check=%d", err, *path, check);

	switch(err){
	case -SYMLINK:
	case -ENOTDIR:
		break;
	case -ENOENT:
		updatecache(err, *path, nil);
		return err;
	default:
		return err;
	}


	redir = 0;
	dynlinks = 0;
	buf = kstrdup(*path);

retry:
	justlinked = 1;
	if(redir++ > 20){
		kprint("%s -> %s - symlink loop\n", *path, buf);
		free(buf);
		return -ELOOP;
	}

	here = buf;

	while((p = strchr(here+1, '/')) != nil){
		*p = 0;
		if((e = fsstat(buf, 1, &ps)) < 0){
			if(justlinked)
				kprint("%s: %s - broken link\n", argv0, *path);
			free(buf);
			updatecache(e, *path, nil);
			return e;
		}
		if(ps.dev == PROCDEV || ps.dev == MISCDEV)
			dynlinks = 1;
		justlinked = 0;

		if((ps.mode & S_IFLNK) == S_IFLNK){
			if((e = deref(&buf, here, p+1)) < 0){
				free(buf);
				updatecache(e, *path, nil);
				return e;
			}
			goto retry;
		}
		*p = '/';
		here = p;
	}

	if(check == Pfull || check == Poptional){
		/*
		 * not a creat(), mkdir(), rename(), link(), or unlink() call
		 */
		if((e = fsstat(buf, 1, &ps)) < 0){
			if(check == Poptional)
				goto success;
			if(justlinked)
				kprint("%s: warning: '%s' -> '%s' broken link\n", argv0, *path, buf);
			free(buf);
			updatecache(e, *path, nil);
			return e;
		}
		if(ps.dev == PROCDEV || ps.dev == MISCDEV)
			dynlinks = 1;
		justlinked = 0;

		if((ps.mode & S_IFLNK) == S_IFLNK){
			if((e = deref(&buf, here, "")) < 0){
				free(buf);
				updatecache(e, *path, nil);
				return e;
			}
			goto retry;
		}
	}
	USED(justlinked);

success:
	cleanname(buf);

	/*
	 * I hit this where apt-get tried to use rmdir() to remove a file. This
	 * gives an ENOTDIR error but for the right reasons. so we end up in
	 * a loop retrying forever.
	 */
	if(strcmp(*path, buf) == 0){
		free(buf);
		return err;
	}
		
	if(! dynlinks)
		updatecache(0, *path, buf);

	free(*path);
	*path = buf;
	return 0;
}
