#include <u.h>
#include <libc.h>
#include <ureg.h>
#include <mp.h>
#include <libsec.h>

#include "dat.h"
#include "fns.h"
#include "linux.h"

enum {
	Qproc,
	Qstat,
	Qcpuinfo,
	Qmeminfo,
	Quptime,
	Qmounts,
	Qfilesystems,
	Qloadavg,
	Qversion,
	Qself,
	Qpid,
	Qcwd,
	Qcmdline,
	Qenviron,
	Qexe,
	Qmountinfo,
	Qroot,
	Qpidstat,
	Qpidstatm,
	Qstatus,
	Qmaps,
	Qfd,
	Qfd1,
	Qtask,
	Qtask1,
	Qplan9,
	Qpathcache,
	Qsyscalls,
	Qrootpath,
};

static struct {
	int mode;
	char *name;
} procdevtab[] = {
	[Qproc]	0555|S_IFDIR,	"proc",
	[Qstat]		0444|S_IFREG,	"stat",
	[Qcpuinfo]	0444|S_IFREG,	"cpuinfo",
	[Qmeminfo]	0444|S_IFREG,	"meminfo",
	[Quptime]	0444|S_IFREG,	"uptime",
	[Qmounts]	0444|S_IFREG,	"mounts",
	[Qfilesystems]	0444|S_IFREG,	"filesystems",		/* exists only to keep apt-get happy */
	[Qloadavg]	0444|S_IFREG,	"loadavg",
	[Qversion]	0444|S_IFREG,	"version",
	[Qself]		0555|S_IFLNK,	"self",
	[Qpid]		0555|S_IFDIR,	"###",
	[Qcwd]			0444|S_IFLNK,	"cwd",
	[Qcmdline]		0444|S_IFREG,	"cmdline",
	[Qenviron]		0444|S_IFREG,	"environ",
	[Qexe]			0444|S_IFLNK,	"exe",
	[Qmountinfo]		0444|S_IFREG,	"mountinfo",	/* exists only to keep apt-get happy */
	[Qroot]			0444|S_IFLNK,	"root",
	[Qpidstat]		0444|S_IFREG,	"stat",
	[Qpidstatm]		0444|S_IFREG,	"statm",
	[Qstatus]		0444|S_IFREG,	"status",
	[Qmaps]			0444|S_IFREG,	"maps",
	[Qfd]			0500|S_IFDIR,	"fd",
	[Qfd1]				0444|S_IFLNK,	"###",
	[Qtask]			0555|S_IFDIR,	"task",
	[Qtask1]			0555|S_IFDIR,	"###",
	[Qplan9]	0755|S_IFDIR,	"plan9",
	[Qpathcache]		0666|S_IFREG,	"pathcache",
	[Qsyscalls]		0444|S_IFREG,	"syscalls",
	[Qrootpath]		0444|S_IFREG,	"rootpath",
};

typedef struct Procfile Procfile;
struct Procfile
{
	Ufile;
	int	q;
	int	pid;
	vlong lastoff;
	char	*data;
	int	ndata;
};

static char mountinfo[] = {
	/* ID parent major:minor rootdir mountpoint... */
	"16 21 0:3 / /proc rw,relatime - proc /proc rw\n"
	"8 23 0:5 / /dev rw,relatime - devtmpfs udev rw,seclabel,size=6084776k,nr_inodes=1521194,mode=755\n"
	"19 18 0:11 / /dev/pts rw,relatime - devpts devpts rw,seclabel,gid=5,mode=620,ptmxmode=000\n"
	"22 21 0:14 / /selinux rw,relatime - selinuxfs none rw\n"
};

static char filesystems[] = {
	"nodev	rootfs\n"
	"nodev	proc\n"
	"	ext4\n"
};

static char *Rootpath = "/";

static int
path2q(char *path, int *ppid, int *pfd)
{
	int i, q, pid, fd;
	char *x;

	q = -ESRCH;
	pid = -1;
	fd = -1;
	path++;
	x = nil;
	for(i = Qproc; i< nelem(procdevtab); i++){
		if(x = strchr(path, '/'))
			*x = 0;
		if(path[0]>='0' && path[0]<='9'){
			switch(i){
			case Qpid:
			case Qtask1:
				pid = atoi(path);
				goto match;
			case Qfd1:
				fd = atoi(path);
				goto match;
			}
		}
		if(strcmp(path, procdevtab[i].name) == 0){
match:
			if(x == nil){
				q = i;
				break;
			}
			if((procdevtab[i].mode & ~0777) == S_IFDIR){
				path = x+1;
				if(i == Qtask1)
					i = Qpid;
			}
			else{
				q = -ENOTDIR;
				break;
			}
		}
		if(x)
			*x = '/';
	}
	if(x)
		*x = '/';
	if(ppid)
		*ppid = pid;
	if(pfd)
		*pfd = fd;
	return q;
}


static int
readlinkproc(char *path, char *buf, int len);

static int
openproc(char *path, int omode, int perm, Ufile **pf)
{
	int q, pid;
	Procfile *f;

	USED(perm);
	if((q = path2q(path, &pid, nil)) < 0)
		return q;
	if((procdevtab[q].mode & S_IFLNK) == S_IFLNK)
		return -SYMLINK;
	if((omode & O_ACCMODE) != O_RDONLY &&
		(procdevtab[q].mode & 0222) == 0)	/* FIXME: should have a real permissions checking func */
		return -EPERM;
	if(q >= Qpid && q < Qplan9){
		qlock(&proctab);
		if(getproc(pid) == nil){
			qunlock(&proctab);
			return -ESRCH;
		}
		qunlock(&proctab);
	}

	f = kmallocz(sizeof(*f), 1);
	f->ref = 1;
	f->omode = omode;
	f->path = kstrdup(path);
	f->fd = -1;
	f->dev = PROCDEV;
	f->q = q;
	f->pid = pid;
	*pf = f;
	return 0;
}

static int
closeproc(Ufile *file)
{
	Procfile *f = (Procfile*)file;

	if(f->data)
		free(f->data);
	return 0;
}

enum {
	SScpu,
	SSswitches,
	SSinterrupts,
	SSsyscalls,
	SSpagefaults,
	SStlbmisses,
	SStlbpurges,
	SSloadavg,
	SSidletime,
	SSintrtime,
	SSmax,
};

static char*
sysstat(ulong *prun, ulong *pidle, ulong *pload)
{
	char buf[1024], *p, *e, *t, *data;
	ulong dt, swtch, user, sys, load;
	static ulong run, idle, intr;
	int n, fd;

	data = nil;
	swtch = user = sys = load = 0;

	dt = (ulong)(((nsec() - boottime) * HZ) / 1000000000LL) - run;
	run += dt;

	n = 0;
	if((fd = open("/dev/sysstat", OREAD)) >= 0){
		n = read(fd, buf, sizeof(buf)-1);
		close(fd);
	}
	if(n > 0){
		buf[n] = 0;
		p = buf;
		while(e = strchr(p, '\n')){
			char *f[SSmax];
			*e = 0;
			if(getfields(p, f, SSmax, 1, "\t ") != SSmax)
				break;

			if(p == buf){
				swtch += atoi(f[SSswitches]);

				idle += (atoi(f[SSidletime]) * dt)/100;
				intr += (atoi(f[SSintrtime]) * dt)/100;

				load = 100-atoi(f[SSidletime]);

				user = run - idle - intr;
				sys = run - user;

				data = ksmprint("cpu %lud %lud %lud %lud %lud %lud %lud\n",
					user, 0UL, sys, idle, 0UL, intr, 0UL);
			}
			t = ksmprint("%scpu%d %lud %lud %lud %lud %lud %lud %lud\n",
				data, atoi(f[SScpu]), user, 0UL, sys, idle, 0UL, intr, 0UL);
			free(data);
			data = t;

			p = e+1;
		}
		t = ksmprint("%sbtime %lud\nctxt %lud\n", data, 
				(ulong)(boottime/1000000000LL), swtch);
		free(data);
		data = t;
	}
	if(prun)
		*prun = run;
	if(pidle)
		*pidle = idle;
	if(pload)
		*pload = load;

	return data;
}

struct linux_sysinfo {
    long uptime;             /* Seconds since boot */
    unsigned long loads[3];  /* 1, 5, and 15 minute load averages */
    unsigned long totalram;  /* Total usable main memory size */
    unsigned long freeram;   /* Available memory size */
    unsigned long sharedram; /* Amount of shared memory */
    unsigned long bufferram; /* Memory used by buffers */
    unsigned long totalswap; /* Total swap space size */
    unsigned long freeswap;  /* swap space still available */
    unsigned short procs;    /* Number of current processes */
    unsigned long totalhigh; /* Total high memory size */
    unsigned long freehigh;  /* Available high memory size */
    unsigned int mem_unit;   /* Memory unit size in bytes */
    char _f[20-2*sizeof(long)-sizeof(int)]; /* Padding for libc5 */
};

int sys_sysinfo(void *buf)
{
	ulong load;
	struct linux_sysinfo *si = buf;

	free(sysstat(nil, nil, &load));

	si->uptime = (nsec() - boottime)/HZ;
	si->loads[0] = load;
	si->loads[1] = 0;
	si->loads[2] = 0;
	si->totalram = 1024*1024*1024;
	si->freeram = si->totalram - 1024*1024*128;
	si->sharedram = 1024*1024*32;
	si->bufferram = 1024*1024*16;
	si->totalswap = 1024*1024*128;
	si->freeswap = 1024*1024*128;
	si->procs = 32;
	si->totalhigh = 1024*1024*128;
	si->freehigh = 1024*1024*128;
	si->mem_unit = PAGESIZE;
	return 0;
}


static char*
procstat(Uproc *p)
{
	if(p->wstate & WEXITED)
		return "Z (zombie)";
 	if(p->wstate & WSTOPPED)
		return "T (stopped)";
	if(p->state == nil)
		return "R (running)";
	return "S (sleeping)";
}

static char*
procname(Uproc *p)
{
	char *s;

	p = getproc(p->pid);
	if(p == nil || p->comm == nil)
		return "";
	if(s = strrchr(p->comm, '/'))
		return s+1;
	return p->comm;
}

static void
gendata(Procfile *f)
{
	char *s, *t;
	int i, n, nproc, nready, ttynum, ttygrp;
	ulong tms[4];
	Uproc *p;

	f->ndata = 0;
	if(s = f->data){
		f->data = nil;
		free(s);
	}
	s = nil;

	if(f->q >= Qpid && f->q < Qplan9){
		ulong vmsize, vmdat, vmlib, vmshr, vmstk, vmexe;

		qlock(&proctab);
		if((p = getproc(f->pid)) == nil){
			qunlock(&proctab);
			return;
		}
		switch(f->q){
		case Qcmdline:
			p = getproc(p->pid);
			if(p == nil || p->comm == nil)
				break;
			i = strlen(p->comm)+1;
			if(i >= p->ncomm-2)
				break;
			f->ndata = p->ncomm-i-2;
			f->data = kmalloc(f->ndata);
			memmove(f->data, p->comm + i, f->ndata);
			qunlock(&proctab);
			return;

		case Qmountinfo:
			f->data = kstrdup(mountinfo);
			f->ndata = strlen(f->data);
			qunlock(&proctab);
			return;

		case Qfilesystems:
			f->data = kstrdup(filesystems);
			f->ndata = strlen(f->data);
			qunlock(&proctab);
			return;

		case Qenviron:
			break;
		case Qpidstat:
			if(proctimes(p, tms) != 0)
				memset(tms, 0, sizeof(tms));
			vmsize = procmemstat(p, nil, nil, nil, nil, nil);
			ttyinfo(p, &ttynum, &ttygrp);

			/* have seen a suicide here once (via ps(1)) though I cannot understand how */
			assert(p != nil);
			assert(p->ureg != nil);
			assert(procname(p) != nil);
			assert(procstat(p) != nil);

			s = ksmprint(
				"%d (%s) %c %d %d %d %d %d %lud %lud "
				"%lud %lud %lud %lud %lud %ld %ld %ld %ld %ld "
				"%ld %lud %lud %ld %lud %lud %lud %lud %lud %lud "
				"%lud %lud %lud %lud %lud %lud %lud %d %d\n",
				p->tid,
				procname(p),
				procstat(p)[0],
				p->ppid, 
				p->pgid,
				p->psid,
				ttynum,	/* tty */
				ttygrp,	/* tty pgrp */
				0UL,	/* flags */
				0UL, 0UL, 0UL, 0UL,	/* pagefault stats */
				tms[0]*HZ / 1000,	/* utime */
				tms[1]*HZ / 1000,	/* stime */
				tms[2]*HZ / 1000,	/* cutime */
				tms[3]*HZ / 1000,	/* cstime */
				20UL,	/* priority */
				0UL,	/* nice */
				0UL,	/* always 0UL */
				0UL,	/* time to next alarm */
				(ulong)(((p->starttime - boottime) * HZ) / 1000000000LL),
				vmsize,	/* vm size in bytes */
				vmsize / PAGESIZE,	/* resident set size in pages */
				4294967295UL,	/* limit of RSS in bytes */
				p->codestart,
				p->codeend,
				p->stackstart,
				p->ureg->sp,	/* SP */
				p->ureg->pc,	/* PC */
				0UL,	/* pending signal mask */
				0UL,	/* blocked signal mask */
				0UL,	/* ignored signal mask */
				0UL,	/* caught signal mask */
				0UL,	/* wait channel address */
				0UL,	/* number of pages swapped */
				0UL,	/* number of children's pages swapped */
				p->exitsignal,
				1,	/* cpu number we last ran on */
				0,	/* RT priority */
				0);	/* scheduling policy */
			break;
		case Qpidstatm:
			vmsize = procmemstat(p, &vmdat, &vmlib, &vmshr, &vmstk, &vmexe);
			s = ksmprint("%lud %lud %lud %lud %lud %lud %lud\n", 
				vmsize/PAGESIZE, vmsize/PAGESIZE, vmshr/PAGESIZE,
				vmexe/PAGESIZE, vmstk/PAGESIZE, vmlib/PAGESIZE, 0UL);
			break;
		case Qstatus:
			s = ksmprint(
				"Name:\t%s\n"
				"State:\t%s\n"
				"Tgid:\t%d\n"
				"Pid:\t%d\n"
				"PPid:\t%d\n"
				"Uid:\t%d\t%d\t%d\t%d\n"
				"Gid:\t%d\t%d\t%d\t%d\n"
				"FDSize:\t%d\n"
				"Threads:\t%d\n",
				procname(p),
				procstat(p),
				p->pid,
				p->tid,
				p->ppid,
				p->uid, p->uid, p->uid, p->uid,
				p->gid, p->gid, p->gid, p->gid,
				MAXFD,
				threadcount(p->pid));
			break;
		case Qmaps:
			break;
		}
		qunlock(&proctab);
	} else {
		ulong run, idle, load;

		nproc = nready = 0;
		qlock(&proctab);
		for(i=0; i<MAXPROC; i++){
			p = getprocn(i);
			if(p == nil)
				continue;
			nproc++;
			if(p->state == nil)
				nready++;
		}
		i = proctab.nextpid;
		qunlock(&proctab);

		switch(f->q){
		case Qstat:
			s = sysstat(nil, nil, nil);
			t = ksmprint(
				"%s"
				"processes %d\n"
				"procs_running %d\n"
				"procs_blocked %d\n",
				s,
				i, 
				nready,
				nproc-nready);
			free(s);
			s = t;
			break;
		case Qcpuinfo:
			break;
		case Qmeminfo:
			break;
		case Quptime:
			free(sysstat(&run, &idle, nil));
			s = ksmprint("%lud.%lud %lud.%lud\n", run/HZ, run%HZ, idle/HZ, idle%HZ);
			break;
		case Qmounts:
			s = mounts();
			break;
		case Qloadavg:
			free(sysstat(nil, nil, &load));
			s = ksmprint("%lud.%lud 0 0 %d/%d %d\n", load/100, load%100, nready, nproc, i);
			break;
		case Qsyscalls:
			s = syscallstats();
			break;
		case Qrootpath:
			s = kstrdup(Rootpath);
			break;
		case Qpathcache:
			s = cacheentries();
			break;
		case Qversion:	/* ps(1) wants this */
			free(sysstat(&run, &idle, nil));
			n = 1;
			if((t = getenv("NPROC")) != nil){
				n = atoi(t);
				free(t);
			}
			s = ksmprint("Linux version 9.9.99-686 #%d SMP %s", n, ctime(boottime/1000000000LL));
			break;

		}
	}

	f->data = s;
	f->ndata = s ? strlen(s) : 0;
}

static vlong
sizeproc(Ufile *file)
{
	Procfile *f = (Procfile*)file;

	if(f->data == nil)
		gendata(f);
	return f->ndata;
}

static int
readproc(Ufile *file, void *buf, int len, vlong off)
{
	Procfile *f = (Procfile*)file;
	int ret;

	if((f->data == nil) || (off != f->lastoff))
		gendata(f);
	ret = 0;
	if(f->data && (off < f->ndata)){
		ret = f->ndata - off;
		if(ret > len)
			ret = len;
		memmove(buf, f->data + off, ret);
		f->lastoff = off + ret;
	}
	return ret;
}

static int
writeproc(Ufile *file, void *buf, int len, vlong)
{
	char *s, *a[8];
	Procfile *f = (Procfile*)file;

	if(f->q != Qpathcache)
		return -EPERM;
	s = buf;
	s[len] = 0;
	if(tokenize(s, a, nelem(a)) < 1)
		return -EINVAL;
	if(flushcache(a[0]) < 0)
		return -ESRCH;
	return 0;
}

static int
readlinkproc(char *path, char *buf, int len)
{
	int n, q, pid, fd;
	char *data;
	Uproc *p;
	Ufile *a;

	if((q = path2q(path, &pid, &fd)) < 0)
		return q;

	if((procdevtab[q].mode & S_IFLNK) != S_IFLNK)
		return -EINVAL;

	data = nil;
	if(q == Qself){
		data = ksmprint("/proc/%d", current->pid);
	}
	else
	if(q >= Qpid){
		qlock(&proctab);
		if((p = getproc(pid)) == nil){
			qunlock(&proctab);
			return -EIO;
		}
		switch(q){
		case Qcwd:
			data = kstrdup(p->cwd);
			break;
		case Qexe:
			data = kstrdup(p->comm);
			break;
		case Qroot:
			if(p->root)
				data = kstrdup(p->root);
			else
				data = kstrdup("/");
			break;
		case Qfd1:
			a = procfdgetfile(p, fd);
			if(a == nil || a->path == nil){
				putfile(a);
				qunlock(&proctab);
				return -ESRCH;
			}
			data = kstrdup(a->path);
			putfile(a);
			break;
		}
		qunlock(&proctab);
	}

	if(data == nil)
		return 0;

	n = strlen(data);
	if(n > len)
		n = len;
	memset(buf, 0, len);
	memmove(buf, data, n);
	free(data);
	return n;
}

static int
readdirproc(Ufile *file, Udirent **pd)
{
	Procfile *f = (Procfile*)file;
	char buf[12];
	Uproc *p;
	Ufile *a;
	int n, i;

	n = 0;
	switch(f->q){
	case Qproc:
		for(i=f->q+1; (procdevtab[i].mode & ~0777) != S_IFDIR; i++){
			if((*pd = newdirent(procdevtab[i].name, procdevtab[i].mode, mkino(Qproc, i))) == nil)
				break;

			pd = &((*pd)->next);
			n++;
		}
		if((*pd = newdirent(procdevtab[Qplan9].name, procdevtab[Qplan9].mode, mkino(Qplan9, 0))) == nil)
			break;
		pd = &((*pd)->next);
		n++;
		/* FALLTHRU */
	case Qtask:
		qlock(&proctab);
		for(i=0; i<MAXPROC; i++){
			p = getprocn(i);
			if(p == nil)
				continue;
			if((f->q == Qproc) && (p->pid != p->tid))
				continue;
			if((f->q == Qtask) && (p->pid != f->pid))
				continue;
			snprint(buf, sizeof(buf), "%d", p->tid);
			if((*pd = newdirent(buf, procdevtab[i].mode, mkino(Qtask, i))) == nil)
				break;
			pd = &((*pd)->next);
			n++;
		}
		qunlock(&proctab);
		break;
	case Qpid:
		if((*pd = newdirent(procdevtab[Qtask].name, procdevtab[Qtask].mode, mkino(Qtask, 0))) == nil)
			break;
		pd = &((*pd)->next);
		n++;
		/* FALLTHRU */
	case Qtask1:
		if((*pd = newdirent(procdevtab[Qfd].name, procdevtab[Qfd].mode, mkino(Qfd, 0))) == nil)
			break;
		pd = &((*pd)->next);
		n++;
		for(i=Qpid+1; (procdevtab[i].mode & ~0777) != S_IFDIR; i++){
			if((*pd = newdirent(procdevtab[i].name, procdevtab[i].mode, mkino(Qpid, i))) == nil)
				break;
			pd = &((*pd)->next);
			n++;
		}
		break;

	case Qfd:
		qlock(&proctab);
		if((p = getproc(f->pid)) == nil){
			qunlock(&proctab);
			break;
		}
		for(i=0; i<MAXFD; i++){
			a = procfdgetfile(p, i);
			if(a == nil || a->path == nil){
				putfile(a);
				continue;
			}
			putfile(a);
			snprint(buf, sizeof(buf), "%d", i);
			if((*pd = newdirent(buf, procdevtab[Qfd1].mode, mkino(Qfd, i))) == nil)
				break;
			pd = &((*pd)->next);
			n++;
		}
		qunlock(&proctab);
		break;
	case Qplan9:
		for(i=Qplan9+1; i < nelem(procdevtab); i++){
			if((*pd = newdirent(procdevtab[i].name, procdevtab[i].mode, mkino(Qplan9, i))) == nil)
				break;
			pd = &((*pd)->next);
			n++;
		}
		break;
	}

	return n;
}

static int
statproc(char *path, int link, Ustat *s)
{
	int q, pid, fd, err;
	Uproc *p;
	Ufile *a;
	err = 0;

	if((q = path2q(path, &pid, &fd)) < 0)
		return q;
	if(! link && (procdevtab[q].mode & S_IFLNK) == S_IFLNK)
		return -SYMLINK;

	s->mode = procdevtab[q].mode;
	s->size = 0;
	s->ino = mkino(PROCDEV, q);
	s->dev = PROCDEV;

	if(q >= Qpid && q < Qplan9){
		qlock(&proctab);
		if((p = getproc(pid)) == nil){
			err = -ENOENT;
			goto out;
		}
		s->uid = p->uid;
		s->gid = p->gid;
		s->atime = s->mtime = s->ctime = p->starttime/1000000000LL;
		switch(q){
		case Qexe:
			if(p->comm == nil){
				err = -EIO;
				goto out;
			}
			break;
		case Qfd1:
			a = procfdgetfile(p, fd);
			if(a == nil || a->path == nil){
				putfile(a);
				err = -ENOENT;
				goto out;
			}
			putfile(a);
			break;
		case Qcwd:
		case Qroot:
		default:
			break;
		}
	}
	if(q == Qself){
		s->uid = current->uid;
		s->gid = current->gid;
		s->atime = s->mtime = s->ctime = boottime/1000000000LL;
	}
	else{
		s->uid = 0;
		s->gid = 0;
		s->atime = s->mtime = s->ctime = boottime/1000000000LL;
	}
out:
	if(q >= Qpid && q < Qplan9)
		qunlock(&proctab);
	return err;
}

static int
fstatproc(Ufile *f, Ustat *s)
{
	return statproc(f->path, 0, s);
};

static int
unlinkproc(char *, int)
{
	return -ENOENT;		/* for the benifit of fs.c, causes a call to retry() */
}

static Udev procdev =
{
	.name = "proc",
	.open = openproc,
	.read = readproc,
	.write = writeproc,
	.size = sizeproc,
	.readlink = readlinkproc,
	.readdir = readdirproc,
	.unlink = unlinkproc,
	.close = closeproc,
	.stat = statproc,
	.fstat = fstatproc,
};

void
procdevinit(char *root)
{
	devtab[PROCDEV] = &procdev;

	Rootpath = root;
	fsmount(&procdev, "/proc");
}
