#include <u.h>
#include <libc.h>
#include <ureg.h>
#include "dat.h"
#include "fns.h"
#include "linux.h"

typedef struct Rfile Rfile;
typedef struct Rpath Rpath;

struct Rfile
{
	Ufile;
	struct
	{
		Dir	*d;
		int	i;
		int	n;
	}		diraux;
};

struct Rpath
{
	Ref;

	Rpath	*hash;

	int		deleted;
	char		str[];
};

typedef struct Hlink Hlink;
typedef struct Hfile Hfile;
struct Hfile {
	Hfile *next;
	char *rpath;
};

struct Hlink {
	Hlink *next;
	Hfile *files;
};

static Rpath	*rpathtab[64];
static QLock	rpathtablk;

static char Fifomagic[] = "#fifo";
static char Linkmagic[] = "#symlink";
static char Charmagic[] = "#chrdev";
static char Blockmagic[] = "#blkdev";
static char Socketmagic[] = "#socket";

static Hlink *Hlroot = nil;

static char *Root;

/*******************************
 * we keep a volatile store of which files are hard linked together.
 * this is only for the benfit of useradd(1) and groupadd(1) which insist on
 * stat(2)ing files and checking the st_rlink after doing a link(2).
 */
void
hllink(char *rold, char *rnew)
{
	Hlink *l;
	Hfile *f;

	/*
	 * linear list is fine, 99.9% of the time the list is ampty anyway */
	for(l = Hlroot; l; l = l->next)
		for(f = l->files; f; f = f->next){
			if(strcmp(f->rpath, rold) == 0){
				f = kmalloc(sizeof(Hfile));
				f->rpath = kstrdup(rnew);
				f->next = l->files;
				l->files = f;
				return;
			}
			if(strcmp(f->rpath, rnew) == 0){
				f = kmalloc(sizeof(Hfile));
				f->rpath = kstrdup(rold);
				f->next = l->files;
				l->files = f;
				return;
			}
		}

	l = kmalloc(sizeof(Hlink));
	l->next = Hlroot;
	l->files = nil;
	Hlroot = l;

	f = kmalloc(sizeof(Hfile));
	f->rpath = kstrdup(rold);
	f->next = l->files;
	l->files = f;

	f = kmalloc(sizeof(Hfile));
	f->rpath = kstrdup(rnew);
	f->next = l->files;
	l->files = f;

}

void
hlunlink(char *rpath)
{
	Hlink *l, **pl;
	Hfile *f, **pf;

	for(l = Hlroot; l; l = l->next){
		pf = &l->files;
		for(f = l->files; f; f = f->next){
			if(strcmp(f->rpath, rpath) == 0){
				*pf = f->next;
				free(f->rpath);
				free(f);
				goto cleanup;
			}
			pf = &f->next;
		}
	}
	return;


cleanup:
	/*
	 * 1/ remove lists of paths which contain only one path.
	 * 2/ remove links which contain no paths
	 */
	pl = &Hlroot;
	for(l = Hlroot; l; l = l->next){
		if(l->files && l->files->next == nil){
			free(l->files->rpath);
			free(l->files);
			l->files = nil;

		}
		if(l->files == nil){
			*pl = l->next;
			free(l);
			break;
		}
		pl = &l->next;
	}	
}

int
hlnlink(char *rpath)
{
	Hlink *l;
	Hfile *f;
	int nlink;

	for(l = Hlroot; l; l = l->next)
		for(f = l->files; f; f = f->next)
			if(strcmp(f->rpath, rpath) == 0)
				goto found;
	if(l == nil)
		return 1;

found:
	nlink = 0;
	for(f = l->files; f; f = f->next)
		nlink++;
	return nlink;
}

static char *
realpath(char *path)
{
	char *p;

	if(path[0] != '/')
		p = kstrdup(path);
	else
		p = ksmprint("%s%s", Root, path);
	trace("	realpath: %s + %s -> %s", Root, path, p);
	return p;

}

static int
readspecial(char *rpath, Dir *d, char *val, int *len)
{
	Dir *xx;
	int t, ret, fd;
	char *p, e[ERRMAX], *buf;

	fd = -1;
	xx = nil;
	buf = nil;

	trace("		readspecial rpath=%s", rpath);
	if(d == nil){
		if((xx = dirstat(rpath)) == nil){
			ret = -ESRCH;
			goto out;
		}
		d = xx;
	}

	if((d->mode & (DMEXCL|DMAPPEND)) != (DMEXCL|DMAPPEND)){
		ret = -EINVAL;
		goto out;
	}

	buf = kmalloc(d->length+1);

	/*
	 * as special files have the exclusive flag set on plan9
	 * we can race trying reading one, so we retry.
	 */
	for(t = 0; (fd = open(rpath, OREAD)) == -1; t++){
		errstr(e, sizeof(e));
		if(t > 10 || strstr(e, "exclusive") == nil){
			ret = mkerror();
			goto out;
		}
		sleep(100);
	}

	if(read(fd, buf, d->length) != d->length){
		ret = mkerror();
		goto out;
	}

	buf[d->length] = 0;
	if((p = strchr(buf, '\n')) != nil)
		*p++ = 0;

	ret = 0;	/* not a special file */

	if(strcmp(buf, Fifomagic) == 0)
		ret = S_IFIFO;
	else if(strcmp(buf, Linkmagic) == 0)
		ret = S_IFLNK;
	else if(strcmp(buf, Charmagic) == 0)
		ret = S_IFCHR;
	else if(strcmp(buf, Blockmagic) == 0)
		ret = S_IFBLK;
	else if(strcmp(buf, Socketmagic) == 0)
		ret = S_IFSOCK;

	/* update the returned length to be the size of the link file
	 * as some programs compare the results from readlink with the
	 * file length returned by stat(2). we put the filename at the
	 * start of the buffer and pad it with zeros
	 */
	if(p && val && *len > 0){
		memset(val, 0, *len);
		strncpy(val, p, *len);
		if((p = strchr(val, '\n')) != nil)
			*p = 0;
		if(d->length < *len)
			*len = d->length;
	}

out:
	free(xx);
	free(buf);
	if(fd != -1)
		close(fd);
	return ret;
}


static Rpath **
rpathent(char *path)
{
	Rpath **prp;

	prp = &rpathtab[hashpath(path) % nelem(rpathtab)];
	while(*prp){
		if(strcmp(path, (*prp)->str) == 0)
			break;
		prp = &((*prp)->hash);
	}
	return prp;
}

static int
ropen(char *path, int omode, int perm, Ufile **pf)
{
	Ufile *f;
	int m, err;
	int omode9, perm9;
	int fd;
	char *rpath;
	Rpath **prp;

	trace("ropen(%s, %#o, %#o, ...)", path, omode, perm);

	omode9 = omode & 3;
	perm9 = (perm & ~current->umask) & 0777;

	rpath = realpath(path);
	if(omode & O_CREAT) {

		if(omode & (O_EXCL | O_TRUNC)){
			if(omode & O_EXCL)
				omode9 |= OEXCL;
			fd = create(rpath, omode9, perm9);
		} else {
			/* try open first to avoid truncating existing the file */
			if((fd = open(rpath, omode9)) < 0)
				fd = create(rpath, omode9, perm9);
		}
		if(fd < 0){
			err = mkerror();
			goto out;
		}
	} else {
		if((m = readspecial(rpath, nil, nil, nil)) > 0){
			switch(m){
			case S_IFLNK:
				err = -SYMLINK;
				break;
			case S_IFSOCK:
			case S_IFCHR:
			case S_IFBLK:
			case S_IFIFO:
			default:
				err = -ENOENT;
				break;
			}
			goto out;
		}

		if(((omode & 3) == O_RDWR) || ((omode & 3) == O_WRONLY))
			if(omode & O_TRUNC)
				omode9 |= OTRUNC;
		if((fd = open(rpath, omode9)) < 0){
			err = mkerror();
			goto out;
		}
	}

	qlock(&rpathtablk);
	prp = rpathent(path);
	if(*prp != nil){
		incref(*prp);
	} else {
		Rpath *rp;

		rp = kmalloc(sizeof(*rp) + strlen(path) + 1);
		rp->ref = 1;
		rp->hash = nil;
		rp->deleted = 0;
		strcpy(rp->str, path);
		*prp = rp;
	}
	qunlock(&rpathtablk);

	f = kmallocz(sizeof(Rfile), 1);
	f->ref = 1;
	f->path = kstrdup(path);
	f->dev = ROOTDEV;
	f->omode = omode;
	f->fd = fd;
	f->off = 0;
	*pf = f;

	err = 0;

out:
	free(rpath);

	return err;
}

static int
rclose(Ufile *f)
{
	Rpath **prp;
	Rfile *file = (Rfile*)f;
	static char rpath[1024];	/* protected by rpathtablk */

	qlock(&rpathtablk);
	prp = rpathent(file->path);
	if(!decref(*prp)){
		Rpath *rp = *prp;
		*prp = rp->hash;
		if(rp->deleted){
			if(fd2path(file->fd, rpath, sizeof(rpath)) == 0)
				remove(rpath);
		}
		free(rp);
	}
	qunlock(&rpathtablk);

	close(file->fd);
	return 0;
}

static int
rread(Ufile *f, void *buf, int len, vlong off)
{
	Rfile *file = (Rfile*)f;
	int ret;

	notifyme(1);
	ret = pread(file->fd, buf, len, off);
	notifyme(0);

	if(ret < 0)
		ret = mkerror();
	return ret;
}

static int
rwrite(Ufile *f, void *buf, int len, vlong off)
{
	Rfile *file = (Rfile*)f;
	int ret;

	notifyme(1);
	ret = pwrite(file->fd, buf, len, off);
	notifyme(0);

	if(ret < 0)
		ret = mkerror();
	return ret;
}

static vlong
rsize(Ufile *f)
{
	Rfile *file = (Rfile*)f;

	return seek(file->fd, 0, 2);
}

static int
raccess(char *path, int mode)
{
	static char omode[] = {
		0,			// ---
		OEXEC,		// --x
		OWRITE,		// -w- 
		ORDWR,		// -wx
		OREAD,		// r--
		OEXEC,		// r-x
		ORDWR,		// rw-
		ORDWR		// rwx
	};

	int err;
	int fd;
	Dir *d;
	char *rpath;
	err = -EINVAL;
	if(mode & ~07)
		return err;

	rpath = realpath(path);
	if((d = dirstat(rpath)) == nil){
		err = mkerror();
		goto out;
	}

	/* ignore the exec bit... firefox gets confused */
	mode &= ~01;
	if((mode == 0) || (d->mode & DMDIR)){
		err = 0;
	} else {
		err = -EACCES;
		if((mode & 01) && ((d->mode & 0111) == 0))
			goto out;
		if((mode & 02) && ((d->mode & 0222) == 0))
			goto out;
		if((mode & 04) && ((d->mode & 0444) == 0))
			goto out;
		if((fd = open(rpath, omode[mode])) >= 0){
			close(fd);
			err = 0;
		}
	}
out:
	free(d);
	free(rpath);
	return err;
}

static ulong
dir2statmode(Dir *d, char *rpath)
{
	int m;
	ulong mode;

	mode = d->mode & 0777;
	if(d->mode & DMDIR)
		return mode | S_IFDIR;
	if(d->type == '|') 
		return mode | S_IFIFO;
	if(d->type == 'H')
		return mode | S_IFBLK;
	if((d->mode & (DMAPPEND|DMEXCL)) == (DMAPPEND|DMEXCL))
		if((m = readspecial(rpath, d, nil, nil)) > 0)
			return mode | m;
	return mode | S_IFREG;
}

static void
dir2ustat(Dir *d, Ustat *s, char *rpath)
{
	s->mode = dir2statmode(d, rpath);
	s->uid = current->uid;
	s->gid = current->gid;
	s->size = d->length;
	s->atime = d->atime;
	s->mtime = d->mtime;
	s->ctime = d->mtime;
	s->ino = d->qid.path;
	s->dev = d->dev;
	s->rdev = 0;
	if(s->mode & S_IFDIR)
		s->nlink = 2;
	else
		s->nlink = hlnlink(rpath);

}

static int
rstat(char *path, int link, Ustat *s)
{
	Dir *d;
	char *rpath;

	rpath = realpath(path);
	if((d = dirstat(rpath)) == nil){
		free(rpath);
		return mkerror();
	}

	dir2ustat(d, s, rpath);
	free(rpath);
	free(d);

	if((s->mode & S_IFLNK) == S_IFLNK && link == 0)
		return -SYMLINK;
	return 0;
}

static int
rfstat(Ufile *f, Ustat *s)
{
	Dir *d;
	char *rpath;

	if((d = dirfstat(f->fd)) == nil)
		return mkerror();

	rpath = realpath(f->path);
	dir2ustat(d, s, rpath);
	free(rpath);
	free(d);
	return 0;
}

static int
rreaddir(Ufile *f, Udirent **pd)
{
	Dir *d;
	char *fullpath, *rpath;
	int i, n, mode;

	seek(f->fd, 0, 0);
	n = dirreadall(f->fd, &d);
	if(n < 0)
		return mkerror();
	for(i=0; i<n; i++){
		fullpath = ksmprint("%s/%s", f->path, d[i].name);
		rpath = realpath(fullpath);
		mode = dir2statmode(&d[i], rpath);
		free(fullpath);
		free(rpath);
		if((*pd = newdirent(d[i].name, mode, d[i].qid.path)) == nil)
			break;
		pd = &((*pd)->next);
	}
	free(d);
	return i;
}

static int
rreadlink(char *path, char *buf, int len)
{
	int mode;
	char *rpath;

	trace("rreadlink(%s)", path);

	rpath = realpath(path);
	mode = readspecial(rpath, nil, buf, &len);
	free(rpath);
	if(mode < 0)
		return mode;
	if((mode & S_IFLNK) != S_IFLNK)
		return EINVAL;
	return len;
}

enum {
	COPYSIZE = 8*1024,
};

static int
copyfile(char *from, char *to)
{
	int err, fromfd, tofd;
	char *buf, *rfrom, *rto;
	Dir *ent, *dir;

	dir = nil;
	buf = nil;
	ent = nil;

	tofd = -1;

	trace("copyfile(%s, %s)", from, to);

	rto = realpath(to);
	rfrom = realpath(from);
	if((fromfd = open(rfrom, OREAD)) < 0){
		err = mkerror();
		goto out;
	}
	if((dir = dirfstat(fromfd)) == nil){
		err = mkerror();
		goto out;
	}
	if((err = open(rto, OREAD)) >= 0){
		close(err);
		err = -EEXIST;
		goto out;
	}
	if(dir->mode & DMDIR){
		int n;
		if((tofd = create(rto, OREAD, dir->mode)) < 0){
			err = mkerror();
			goto out;
		}
		close(tofd);
		tofd = -1;
		while((n = dirread(fromfd, &ent)) > 0){
			int i;

			for(i=0; i<n; i++){
				char *froment, *toent;

				froment = allocpath(from, nil, ent[i].name);
				toent = allocpath(to, nil, ent[i].name);
				err = copyfile(froment, toent);
				free(froment);
				free(toent);

				if(err < 0)
					goto out;
			}
			free(ent); ent = nil;
		}
	} else {
		if((tofd = create(rto, OWRITE, dir->mode)) < 0){
			err = mkerror();
			goto out;
		}
		buf = kmalloc(COPYSIZE);
		for(;;){
			err = read(fromfd, buf, COPYSIZE);
			if(err == 0)
				break;
			if(err < 0){
				err = mkerror();
				goto out;
			}
			if(write(tofd, buf, err) != err){
				err = mkerror();
				goto out;
			}
		}
	}

	err = 0;
out:
	free(ent);
	free(dir);
	free(buf);
	close(fromfd);
	close(tofd);
	return err;
}

static int
removefile(char *path)
{
	int err;
	int n;
	Dir *d;
	int fd;
	char *rpath;

	trace("removefile(%s)", path);

	rpath = realpath(path);
	if((d = dirstat(rpath)) == nil){
		free(rpath);
		return mkerror();
	}
	if(remove(rpath) == 0){
		free(rpath);
		free(d);
		return 0;
	}
	if((d->mode & DMDIR) == 0){
		free(rpath);
		free(d);
		return mkerror();
	}
	free(d);

	if((fd = open(rpath, OREAD)) < 0){
		free(rpath);
		return mkerror();
	}
	free(rpath);

	err = 0;
	d = nil;
	while((n = dirread(fd, &d)) > 0){
		char *t;
		int i;

		for(i=0; i<n; i++){
			t = allocpath(path, nil, d[i].name);
			err = removefile(t);
			free(t);

			if(err < 0)
				break;
		}
		free(d); d = nil;

		if(err < 0)
			break;
	}
	close(fd);
	if(err < 0)
		return err;
	if(n < 0)
		return mkerror();
	if(remove(rpath) < 0)
		return mkerror();
	return 0;
}

static int
rrename(char *from, char *to)
{
	int err;
	char *x, *y, *t, *tmp, *rfrom, *rto, *rtmp;

	trace("rrename(%s, %s)", from, to);

	/*
	 * POSIX.1, section 5.5.3.2: 
	 * "If the _old_ argument and the _new_ argument both refer to links 
	 * to the same existing file, the rename() function shall return 
	 * successfully and perform no other action."
	 */
	if(strcmp(from, to) == 0)
		return 0;

	x = nil;
	for(t=from; *t; t++){
		if(*t == '/')
			x = t;
	}
	y = nil;
	for(t=to; *t; t++){
		if(*t == '/')
			y = t;
	}

	rto = realpath(to);
	rfrom = realpath(from);

	if(x && y){
		char *e;

		e = nil;
		*x = 0; *y = 0;
		if(strcmp(from, to) == 0)
			e = &y[1];
		*x = '/'; *y = '/';

		if(e != nil){
			Dir d;

			nulldir(&d);
			d.name = e;

			remove(rto);
			free(rto);

			err = dirwstat(rfrom, &d);
			free(rfrom);
			if(err < 0)
				return mkerror();

			return 0;
		}
	}

	tmp = ksmprint("%s%d%d.tmp", to, current->pid, current->tid);
	rtmp = realpath(tmp);

	if((err = copyfile(from, tmp)) == 0){
		Dir d;

		nulldir(&d);
		d.name = &y[1];

		remove(rto);
		if(dirwstat(rtmp, &d) < 0)
			err = mkerror();
		else
			removefile(from);
	}
	if(err != 0)
		removefile(tmp);
	free(tmp);
	free(rto);
	free(rtmp);
	free(rfrom);
	return err;
}

static int
rmkdir(char *path, int mode)
{
	char *rpath;
	int err, fd, mode9;

	trace("rmkdir(%s, %#o)", path, mode);

	rpath = realpath(path);
	mode9 = DMDIR | ((mode & ~current->umask) & 0777);
	if((fd = create(rpath, OREAD|OEXCL, mode9)) < 0){
		err = mkerror();
		goto out;
	}
	close(fd);
	err = 0;

out:
	free(rpath);
	return err;
}

static void
combinedir(Dir *ndir, Dir *odir)
{
	if(ndir->mode != ~0)
		ndir->mode = (odir->mode & ~0777) | (ndir->mode & 0777);
}

static int
uwstat(char *path, Dir *ndir, int link)
{
	int err;
	Dir *dir;
	char *rpath;

	trace("uwstat(%s, %p, %d)", path, ndir, link);

	err = 0;
	rpath = realpath(path);
	if((dir = dirstat(rpath)) == nil){
		err = mkerror();
		goto out;
	}

	combinedir(ndir, dir);

	if(dirwstat(rpath, ndir) < 0)
		err = mkerror();
out:
	free(rpath);
	free(dir);
	return err;	
}

static int
uwfstat(Ufile *f, Dir *ndir)
{
	int err;
	Dir *dir;

	err = 0;
	if((dir = dirfstat(f->fd)) == nil){
		err = mkerror();
		goto out;
	}
	combinedir(ndir, dir);
	if(dirfwstat(f->fd, ndir) < 0)
		err = mkerror();
out:
	free(dir);
	return err;
}

static int
rutime(char *path, long atime, long mtime)
{
	Dir ndir;

	trace("rutime(%s, %ld, %ld)", path, atime, mtime);

	nulldir(&ndir);
	ndir.atime = atime;
	ndir.mtime = mtime;
	return uwstat(path, &ndir, 1);
}

static int
rchmod(char *path, int mode, int link)
{
	Dir ndir;

	trace("rchmod(%s, %#o)", path, mode);

	nulldir(&ndir);
	ndir.mode = mode;
	return uwstat(path, &ndir, link);
}

static int
rchown(char *path, int uid, int gid, int link)
{
	Ustat s;

	USED(uid);
	USED(gid);

	/* FIXME, just return the right errorcode for now */
	return fsstat(path, link, &s);
}

static int
rtruncate(char *path, vlong size)
{
	Dir ndir;

	trace("rtruncate(%s, %lld)", path, size);

	nulldir(&ndir);
	ndir.length = size;

	return uwstat(path, &ndir, 0);
}

static int
rfchmod(Ufile *f, int mode)
{
	Dir ndir;

	nulldir(&ndir);
	ndir.mode = mode;
	return uwfstat(f, &ndir);
}

static int
rfchown(Ufile *f, int uid, int gid)
{
	USED(f);
	USED(uid);
	USED(gid);

	return 0;
}

static int
rftruncate(Ufile *f, vlong size)
{
	Dir ndir;

	nulldir(&ndir);
	ndir.length = size;
	return uwfstat(f, &ndir);
}

static int
runlink(char *path, int rmdir)
{
	int err;
	Dir *dir;
	char *rpath;
	Rpath **prp;

	trace("runlink(%s, %d)", path, rmdir);

	rpath = realpath(path);
	if((dir = dirstat(rpath)) == nil){
		err = mkerror();
		goto out;
	}

	if(rmdir){
		if((dir->mode & DMDIR) == 0){
			err = -ENOTDIR;
			goto out;
		}
	} else {
		if(dir->mode & DMDIR){
			err = -EISDIR;
			goto out;
		}
	}

	qlock(&rpathtablk);
	hlunlink(rpath);
	prp = rpathent(path);
	if(*prp){
		Dir ndir;
		char *tmp, *tail;

		if((tail = strrchr(rpath, '/')) != nil)
			tail++;
		else
			tail = rpath;
		tmp = ksmprint(".%s.%d.deleted", tail, current->kpid);
		nulldir(&ndir);
		ndir.name = tmp;
		trace("runlink: file %s still in use renaming to -> %s", path, tmp);
		if(dirwstat(rpath, &ndir) < 0){
			qunlock(&rpathtablk);
			err = mkerror();
			free(tmp);
			goto out;
		}
		free(tmp);
		(*prp)->deleted = 1;
		qunlock(&rpathtablk);

	} else {
		qunlock(&rpathtablk);
		if(remove(rpath) < 0){
			err = mkerror();
			goto out;
		}
	}
	err = 0;
out:
	free(dir);
	free(rpath);

	return err;
}

static int
rlink(char *old, char *new, int sym)
{
	int err, fd;
	char *rold, *rnew;

	trace("rlink(%s, %s, %d)", old, new, sym);

	/* new must not exist */
	rnew = realpath(new);
	rold = realpath(old);
	if((fd = open(rnew, OREAD)) >= 0){
		close(fd);
		err = -EEXIST;
		goto out;
	}

	if(sym == 0){
		/* old must exist */
		if((fd = open(rold, OREAD)) < 0){
			err = mkerror();
			goto out;
		}
		close(fd);
		qlock(&rpathtablk);
		hllink(rold, rnew);
		qunlock(&rpathtablk);
		err = copyfile(old, new);
		goto out;
	}

	/* try to create the link, will fail if already exists */
	if((fd = create(rnew, OWRITE|OEXCL, DMAPPEND|DMEXCL|0755)) < 0){
		err = mkerror();
		goto out;
	}

	fprint(fd, "%s\n%s\n", Linkmagic, old);
	close(fd);
	err = 0;
out:
	free(rold);
	free(rnew);
	return err;
}

static int
rmknod(char *path, int mode, int dev)
{
	char *rpath;
	int fd, mode9, err;

	trace("rmknod(%s, %d, 0x%x)", path, mode, dev);

	mode9 = mode & 0777;
	if((mode & S_IFREG) != S_IFREG)
		mode9 |= DMAPPEND|DMEXCL;

	err = 0;
	rpath = realpath(path);
	if((fd = create(rpath, OWRITE, mode9)) < 0)
		return mkerror();

	if((mode & S_IFBLK) == S_IFBLK){
		if(fprint(fd, "%s\n0x%x\n", Blockmagic, dev) < 0){
			err = mkerror();
			goto out;
		}
	}
	else
	if((mode & S_IFCHR) == S_IFCHR){
		if(fprint(fd, "%s\n0x%x\n", Charmagic, dev) < 0){
			err = mkerror();
			goto out;
		}
	}
	else
	if((mode & S_IFIFO) == S_IFIFO){
		if(fprint(fd, "%s\n", Fifomagic) < 0){
			err = mkerror();
			goto out;
		}
	}
	else
	if((mode & S_IFSOCK) == S_IFSOCK){
		if(fprint(fd, "%s\n", Socketmagic) < 0){
			err = mkerror();
			goto out;
		}
	}
	else
	if((mode & S_IFREG) == S_IFREG){
		goto out;
	}
	else
		err = EINVAL;

out:
	free(rpath);
	close(fd);
	return err;
}


static Udev rootdev = 
{
	.name = "root",
	.open = ropen,
	.access = raccess,
	.stat = rstat,
	.link = rlink,
	.unlink = runlink,
	.rename = rrename,
	.mkdir = rmkdir,
	.utime = rutime,
	.chmod = rchmod,
	.chown = rchown,
	.truncate = rtruncate,
	.mknod = rmknod,

	.read = rread,
	.write = rwrite,
	.size = rsize,
	.close = rclose,

	.fstat = rfstat,
	.readdir = rreaddir,
	.readlink = rreadlink,

	.fchmod = rfchmod,
	.fchown = rfchown,
	.ftruncate = rftruncate,
};

void rootdevinit(char *root)
{
	Root = root;
	devtab[ROOTDEV] = &rootdev;
	fsmount(&rootdev, "");
}
