#include <u.h>
#include <libc.h>
#include <ureg.h>
#include <ip.h>
#include "dat.h"
#include "fns.h"
#include "linux.h"
#include "socket.h"

typedef struct Socket Socket;
typedef struct Connectproc Connectproc;
typedef struct Listenproc Listenproc;

enum {
	Ctlsize = 128,
};

union sockaddr_all {
	struct sockaddr sa;
	struct sockaddr_in sin;
	struct sockaddr_un sun;
};

struct Socket {
	Ufile;

	int		family;
	int		stype;
	int		protocol;

	int		other;
	char		net[128];
	char		name[Ctlsize];

	int		naddr;
	uchar		addr[sizeof(union sockaddr_all)];

	void		*bufproc;
	Connectproc	*connectproc;
	Listenproc	*listenproc;

	int		connected;
	int		error;

	Socket		*next;
};

struct Connectproc {
	Ref;
	QLock;
	Socket 	*sock;
	int	notefd;
	Uwaitq	wq;
	char	str[Ctlsize];
};

struct Listenproc {
	Ref;
	QLock;
	Socket	*sock;
	int		notefd;
	Uwaitq	wq;
	Socket	*q;
	char		str[Ctlsize];
};

static char*
srvname(char *npath, char *path, int len)
{
	char *p;

	p = strrchr(path, '/');
	if(p == 0)
		p = path;
	else
		p++;
	snprint(npath, len, "/srv/UD.%s", p);
	return npath;
}

static int
srvunixsock(int fd, char *path)
{
	int ret;
	int sfd;
	char buf[8+Ctlsize+1];

	sfd = -1;
	ret = -1;
	if(fd < 0)
		goto out;
	srvname(buf, path, sizeof(buf));
	remove(buf);
	if((sfd = create(buf, OWRITE, 0666)) < 0)
		goto out;
	sprint(buf, "%d", fd);
	if(write(sfd, buf, strlen(buf)) < 0)
		goto out;
	ret = 0;
out:
	if(sfd >= 0)
		close(sfd);
	return ret;
}

static void
unsrvunixsock(char *path)
{
	char buf[8+Ctlsize+1];

	srvname(buf, path, sizeof(buf));
	remove(buf);
}

static Socket*
allocsock(int family, int stype, int protocol)
{
	Socket *sock;

	sock = kmallocz(sizeof(*sock), 1);
	sock->family = family;
	sock->stype = stype;
	sock->protocol = protocol;
	sock->fd = -1;
	sock->other = -1;
	sock->ref = 1;
	sock->dev = SOCKDEV;
	sock->omode = O_RDWR;

	return sock;
}

static char *
sockfam2str(int n)
{
	static char buf[32];
	char *tab[] = {
		"NONE", "UNIX", "INET", "AX25", "IPX", "APPLETALK", "NETROM",
		"BRIDGE", "ATM PVC", "X25", "INET6", "ROSE", "DECNET", "NETBEUI",
		"SECURITY", "KEY", "NETLINK", "PACKET", "ASH", "ECONET", "ATMSVC",
		"RDS", "SNA", "IRDA", "PPPOX", "WANPIPE", "LLC", "CAN", "TIPC",
		"BLUETOOTH", "IUCV", "RXRPC", "ISDN", "PHONET", "IEEE802154",
	};

	if(n >= 0 && n < nelem(tab) && tab[n])
		return tab[n];
	snprint(buf, sizeof(buf), "(%d)", n);
	return buf;
}

static char *
socktype2str(int type)
{
	int n, len;
	static char buf[64];
	char *tab[] = {
		nil, "DGRAM", "STREAM", "RAW", "RDM", "SEQPACKET", "DCCP", nil,
		nil, nil, "PACKET"
	};
	n = type & 0xff;

	if(n >= 0 && n < nelem(tab) && tab[n])
		snprint(buf, sizeof(buf), "%s", tab[n]);
	else
		snprint(buf, sizeof(buf), "(%x)", n & 0xff);

	len = strlen(buf);
	snprint(buf+len, sizeof(buf)-len, "%s", ((type & SOCK_CLOEXEC)? "|SOCK_CLOEXEC": ""));

	len = strlen(buf);
	snprint(buf+len, sizeof(buf)-len, "%s", ((type & SOCK_NONBLOCK)? "|SOCK_NONBLOCK": ""));
	return buf;
}

static char *
socknetlink2str(int n)
{
	static char buf[32];
	char *tab[] = {
		"ROUTE", "UNUSED", "USERSOCK", "FIREWALL", "INET_DIAG", "NFLOG",
		"XFRM", "SELINUX", "ISCSI", "AUDIT", "FIB_LOOKUP", "CONNECTOR",
		"NETFILTER", "IP6_FW", "DNRTMSG", "KOBJECT_UEVENT", "GENERIC",
		"NETLINK_DM", "SCSITRANSPORT", "ECRYPTFS",
	};
	if(n >= 0 && n < nelem(tab) && tab[n])
		return tab[n];
	snprint(buf, sizeof(buf), "(%d)", n);
	return buf;
}

static int
newsock(int family, int stype, int protocol)
{
	Socket *sock;
	char *net;
	char buf[Ctlsize];
	int pfd[2];
	int cfd, dfd;
	int n;
	int err;

	trace("newsock(AF_%s, SOCK_%s, %x)", sockfam2str(family), socktype2str(stype), protocol);

	err = -EINVAL;
	switch(family){
	case AF_INET:
	case AF_INET6:
		switch(stype & 0xff){
		case SOCK_DGRAM:
			net = "udp";
			break;
		case SOCK_STREAM:
			net = "tcp";
			break;
		default:
			kprint("socket(AF_%s,SOCK_%s,%d) not supported\n",
				sockfam2str(family), socktype2str(stype & 0xff), protocol);
			return err;
		}
		break;
	case AF_UNIX:
		net = nil;
		break;
	default:
		kprint("socket(AF_%s,SOCK_%s,%d) not supported\n",
			sockfam2str(family), socktype2str(stype & 0xff), protocol);
		return err;
	}

	sock = allocsock(family, stype, protocol);
	if(stype & SOCK_NONBLOCK)
		sock->omode |= O_NONBLOCK;

	cfd = -1;
	switch(family){
	case AF_INET:
		snprint(buf, sizeof(buf), "/net/%s/clone", net);
		if((cfd = open(buf, ORDWR)) < 0){
			err = mkerror();
			goto errout;
		}
		n = read(cfd, buf, sizeof(buf)-1);
		if(n < 0)
			err = mkerror();
		if(n <= 0)
			goto errout;
		buf[n] = 0;
		n = atoi(buf);
		snprint(buf, sizeof(buf), "/net/%s/%d/data", net, n);
		if((dfd = open(buf, ORDWR)) < 0){
			err = mkerror();
			goto errout;
		}
		close(cfd);
		sock->fd = dfd;
		snprint(sock->net, sizeof(sock->net), "/net/%s", net);
		snprint(sock->name, sizeof(sock->name), "%s/%d", sock->net, n);
		break;
	case AF_UNIX:
		if(pipe(pfd) < 0){
			err = mkerror();
			goto errout;
		}
		sock->other = pfd[1];
		sock->fd = pfd[0];
		break;
	case AF_INET6:
		kprint("socket(AF_%s,SOCK_%s,%d) not supported\n",
			sockfam2str(family), socktype2str(stype & 0xff), protocol);
		return -EINVAL;
	default:
		sysfatal("socket: unknown family, cannot happen");
	}

	if(stype & SOCK_CLOEXEC)
		return newfd(sock, FD_CLOEXEC);
	else
		return newfd(sock, 0);

errout:
	close(cfd);
	free(sock);
	return err;
}

static void
freeconnectproc(Connectproc *cp)
{
	if(cp == nil)
		return;
	qlock(cp);
	cp->sock = nil;
	if(decref(cp)){
		write(cp->notefd, "interrupt", 9);
		qunlock(cp);
		return;
	}
	qunlock(cp);
	close(cp->notefd);
	free(cp);
}

static void
freelistenproc(Listenproc *lp)
{
	Socket *q;

	if(lp == nil)
		return;
	qlock(lp);
	lp->sock = nil;
	if(decref(lp)){
		write(lp->notefd, "interrupt", 9);
		qunlock(lp);
		return;
	}
	while(q = lp->q){
		lp->q = q->next;
		putfile(q);
	}
	qunlock(lp);
	close(lp->notefd);
	free(lp);
}

static int
closesock(Ufile *file)
{
	Socket *sock = (Socket*)file;

	close(sock->fd);
	close(sock->other);
	freebufproc(sock->bufproc);
	freeconnectproc(sock->connectproc);
	freelistenproc(sock->listenproc);
	return 0;
}


static void
connectproc(void *aux)
{
	int fd, cfd, other;
	char buf[Ctlsize], tmp[8+Ctlsize+1];
	Connectproc *cp;
	Socket *sock;
	int err;

	cp =  (Connectproc*)aux;
	qlock(cp);
	if((sock = cp->sock) == nil)
		goto out;

	snprint(buf, sizeof(buf), "connectproc() %s", cp->str); 
	setprocname(buf);

	err = 0;
	switch(sock->family){
	case AF_UNIX:
		fd = sock->fd;
		other = sock->other;
		qunlock(cp);

		err = -ECONNREFUSED;
		srvname(tmp, cp->str, sizeof(buf));
		if((cfd = open(tmp, ORDWR)) < 0)
			break;

		memset(buf, 0, sizeof(buf));
		snprint(buf, sizeof(buf), "linuxemu.%d.%lux", getpid(), (ulong)sock);
		if(srvunixsock(other, buf) < 0){
			close(cfd);
			break;
		}

		/*
		 * write Ctrlsize-1 bytes so concurrent writes will not be merged together as
		 * Ctrlsize-1 is the size used in read(). see /sys/src/ape/lib/bsd/accept.c:87
		 * this should be fixed in ape's connect() as well.
		 */
		if(write(cfd, buf, sizeof(buf)-1) != sizeof(buf)-1){
			close(cfd);
			unsrvunixsock(buf);
			break;
		}
		close(cfd);
		if((read(fd, tmp, strlen(buf)) != strlen(buf)) || memcmp(buf, tmp, strlen(buf))){
			unsrvunixsock(buf);
			break;
		}
		unsrvunixsock(buf);
		err = 0;
		break;

	case AF_INET:
		snprint(buf, sizeof(buf), "%s/ctl", sock->name);
		qunlock(cp);
		if((cfd = open(buf, ORDWR)) < 0){
			err = mkerror();
			break;
		}

		if(fprint(cfd, "connect %s", cp->str) < 0)
			err = mkerror();
		close(cfd);
		break;
	default:
		sysfatal("connect(%s) cannot happen", sockfam2str(sock->family));
	}

	qlock(cp);
	if((sock = cp->sock) == nil)
		goto out;
	if(err == 0){
		close(sock->other);
		sock->other = -1;
		sock->connected = 1;
	}
	sock->error = err;
out:
	wakeq(&cp->wq, MAXPROC);
	qunlock(cp);
	freeconnectproc(cp);
}

static int
sockaddr2str(Socket *sock, void *addr, int addrlen, char *buf, int nbuf)
{
	struct sockaddr_in *sin;
	struct sockaddr_un *sun;

	switch(sock->family){
	case AF_INET:
		sin = addr;
		if(addrlen < sizeof(struct sockaddr_in))
			break;
		if(sin->sin_family != sock->family)
			break;
		snprint(buf, nbuf, "%V!%d", (uchar *)&sin->sin_addr.s_addr, nhgets(&sin->sin_port));
		trace("	sockaddr2str -> inet: %s\n", buf);
		return 0;

	case AF_UNIX:
		sun = addr;
		if(addrlen >= sizeof(struct sockaddr_un))
			break;
		if(sun->sun_family != sock->family)
			break;
		trace("	sockaddr2str -> unix: %d %d '%s'", addrlen, sun->sun_family, sun->sun_path);
		if(addrlen >= nbuf)
			addrlen = nbuf-1;
		memmove(buf, sun->sun_path, addrlen);
		buf[addrlen] = 0;
		trace("	sockaddr2str -> unix: %s\n", buf);
		return 0;
	default:
		sysfatal("sockaddr2str(%s)", sockfam2str(sock->family));
	}

	return -EINVAL;
}

static int
connectsock(Socket *sock, uchar *addr, int addrlen)
{
	Connectproc *cp;
	int err;
	char buf[Ctlsize];
	int pid;

	if(sock->connected)
		return -EISCONN;
	if(sock->connectproc)
		return -EALREADY;

	if((err = sockaddr2str(sock, addr, addrlen, buf, sizeof(buf))) < 0)
		return err;

	cp = kmallocz(sizeof(*cp), 1);
	cp->ref = 2;
	cp->sock = sock;
	strncpy(cp->str, buf, sizeof(cp->str));

	qlock(cp);
	sock->error = 0;
	if((pid = procfork(connectproc, cp, 0)) < 0){
		qunlock(cp);
		free(cp);
		return mkerror();
	}
	snprint(buf, sizeof(buf), "/proc/%d/note", pid);
	cp->notefd = open(buf, OWRITE);

	if(addrlen > sizeof(sock->addr))
		addrlen = sizeof(sock->addr);
	sock->naddr = addrlen;
	memmove(sock->addr, addr, addrlen);

	/*
	 * glibc's gethostbyname() fails if we return EINPROGRESS
	 * when it creates a UDP socket for DNS lookups. It doesn't
	 * make sense to return EINPROGRESS as a UDP connect is a nop
	 * really.
	 */
	if(sock->omode & O_NONBLOCK)
	if((sock->stype & 0xff) == SOCK_STREAM)
	if(sock->family == AF_INET){
		trace("nonblocking, family=af_in type=sock_stream");
		sock->connectproc = cp;
		qunlock(cp);
		return -EINPROGRESS;
	}

	if((err = sleepq(&cp->wq, cp, 1)) == 0)
		err = sock->error;
	qunlock(cp);
	freeconnectproc(cp);

	return err;
}

static int
shutdownsock(Socket *sock, int how)
{
	USED(how);

	freebufproc(sock->bufproc);
	sock->bufproc = nil;
	freeconnectproc(sock->connectproc);
	sock->connectproc = nil;
	freelistenproc(sock->listenproc);
	sock->listenproc = nil;
	close(sock->fd);
	sock->fd = -1;
	sock->connected = 0;

	return 0;
}

static int 
bindsock(Socket *sock, void *addr, int addrlen)
{
	int port;
	int cfd;
	char buf[Ctlsize];
	struct sockaddr_in *sin;

	switch(sock->family){
	case AF_UNIX:
		break;
	case AF_INET:
		if(addrlen < sizeof(struct sockaddr_in))
			return -EINVAL;
		sin = addr;
		port = nhgets(&sin->sin_port);

		snprint(buf, sizeof(buf), "%s/ctl", sock->name);
		if((cfd = open(buf, ORDWR)) < 0)
			return mkerror();
		if((fprint(cfd, "announce %d", port) < 0) || (fprint(cfd, "bind %d", port) < 0)){
			close(cfd);
			return mkerror();
		}
		close(cfd);
		break;
	default:
		fprint(2, "bind(%s)", sockfam2str(sock->family));
		return -EINVAL;
	}

	if(addrlen > sizeof(sock->addr))
		addrlen = sizeof(sock->addr);
	sock->naddr = addrlen;
	memmove(sock->addr, addr, addrlen);

	return 0;
}

static int
getsockaddr(Socket *sock, int remote, void *addr, int len)
{
	struct sockaddr_in *sin;
	char *p, buf[Ctlsize];
	uchar ipv6[IPaddrlen];
	int fd, ip, n;

	switch(sock->family){
	case AF_UNIX:
		if(len < sock->naddr)
			break;
		memmove(addr, sock->addr, sock->naddr);
		return sock->naddr;

	case AF_INET:
		if(len < sizeof(struct sockaddr_in))
			break;

		if(remote)
			snprint(buf, sizeof(buf), "%s/remote", sock->name);
		else
			snprint(buf, sizeof(buf), "%s/local", sock->name);
		if((fd = open(buf, OREAD)) < 0)
			return mkerror();
		if((n = read(fd, buf, sizeof(buf)-1)) < 0){
			close(fd);
			return mkerror();
		}
		close(fd);
		buf[n-1] = 0;

		trace("getsockaddr(): %s", buf);

		sin = addr;
		memset(sin, sizeof(struct sockaddr_in), 0);

		sin->sin_family = sock->family;

		if((p = strrchr(buf, '!')) != nil)
			hnputs(&sin->sin_port, atoi(p+1));

		ip = parseip(ipv6, buf);
		memcpy(&sin->sin_addr.s_addr, &ip, sizeof(sin->sin_addr.s_addr));

		return sizeof(struct sockaddr_in);
	default:
		sysfatal("getsockaddr(%s)", sockfam2str(sock->family));
	}


	return -EINVAL;
}

static void
listenproc(void *aux)
{
	Listenproc *lp;
	Socket *sock, *q;
	char buf[Ctlsize], tmp[8+Ctlsize+1];
	int cfd, fd, n;

	lp = (Listenproc*)aux;
	qlock(lp);
	if((sock = lp->sock) == nil)
		goto out;

	snprint(buf, sizeof(buf), "listenproc() %s", lp->str);
	setprocname(buf);

	for(;;){
		n = 0;
		cfd = -1;
		switch(sock->family){
		case AF_UNIX:
			srvunixsock(sock->other, lp->str);
			close(sock->other);
			sock->other = -1;
			fd = sock->fd;
			qunlock(lp);
			n = read(fd, buf, sizeof(buf)-1);
			qlock(lp);
			break;

		default:
			snprint(buf, sizeof(buf), "%s/listen", sock->name);
			qunlock(lp);
			if((cfd = open(buf, ORDWR)) >= 0)
				n = read(cfd, buf, sizeof(buf)-1);
			qlock(lp);
			if(n <= 0)
				close(cfd);
		}
		if(n <= 0)
			break;
		buf[n] = 0;

		if((sock = lp->sock) == nil){
			close(cfd);
			break;
		}

		switch(sock->family){
		case AF_UNIX:
			srvname(tmp, buf, sizeof(tmp));
			if((fd = open(tmp, ORDWR)) < 0)
				break;
			unsrvunixsock(buf);
			if(write(fd, buf, strlen(buf)) != strlen(buf)){
				close(fd);
				fd = -1;
			}
			buf[0] = 0;
			break;

		default:
			n = atoi(buf);
			snprint(buf, sizeof(buf), "%s/%d", sock->net, n);
			snprint(tmp, sizeof(tmp), "%s/data", buf);
			fd = open(tmp, ORDWR);
			close(cfd);
			break;
		}

		if(fd < 0)
			continue;

		q = allocsock(sock->family, sock->stype, sock->protocol);
		strncpy(q->net, sock->net, sizeof(q->net));
		strncpy(q->name, buf, sizeof(q->name));

		if(sock->family == AF_UNIX){
			memmove(q->addr, sock->addr, q->naddr = sock->naddr);
		} else {
			q->naddr = getsockaddr(q, 0, q->addr, sizeof(q->addr));
		}

		q->fd = fd;
		q->connected = 1;
		q->next = lp->q;
		lp->q = q;
		wakeq(&lp->wq, MAXPROC);
	}

	if(sock->family == AF_UNIX)
		unsrvunixsock(lp->str);
out:
	wakeq(&lp->wq, MAXPROC);
	qunlock(lp);
	freelistenproc(lp);
}


static int
listensock(Socket *sock)
{
	Listenproc *lp;
	int pid, err;
	char buf[Ctlsize];

	trace("listensock()");

	if(sock->listenproc)
		return 0;
	if((err = sockaddr2str(sock, sock->addr, sock->naddr, buf, sizeof(buf))) < 0)
		return err;

	lp = kmallocz(sizeof(*lp), 1);
	lp->ref = 2;
	lp->sock = sock;
	strncpy(lp->str, buf, sizeof(lp->str));

	qlock(lp);
	if((pid = procfork(listenproc, lp, 0)) < 0){
		qunlock(lp);
		free(lp);
		return mkerror();
	}
	snprint(buf, sizeof(buf), "/proc/%d/note", pid);
	lp->notefd = open(buf, OWRITE);
	sock->listenproc = lp;
	qunlock(lp);

	return 0;
}

static int
getsockname(Socket *sock, uchar *addr, int *paddrlen)
{
	int ret;

	trace("getsockname(%p, %p, %p (%x))", sock, addr, paddrlen, paddrlen ? *paddrlen : 0);

	if(addr == nil || paddrlen == nil)
		return -EINVAL;

	ret = sock->naddr;
	memmove(addr, sock->addr, ret);
	*paddrlen = ret;

	return ret;
}

static int
getpeername(Socket *sock, uchar *addr, int *paddrlen)
{
	int ret;

	trace("getpeername(%p, %p, %p (%x))", sock, addr, paddrlen, paddrlen ? *paddrlen : 0);

	if(addr == nil || paddrlen == nil)
		return -EINVAL;

	if((ret = getsockaddr(sock, 1, addr, *paddrlen)) > 0)
		*paddrlen = ret;
	return ret;
}

static int
acceptsock(Socket *sock, uchar *addr, int *paddrlen)
{
	Listenproc *lp;
	Socket *nsock;
	int err;

	trace("acceptsock(%p, %p, %p (%x))", sock, addr, paddrlen, paddrlen ? *paddrlen : 0);

	if((lp = sock->listenproc) == nil)
		return -EINVAL;

	qlock(lp);
	for(;;){
		if(nsock = lp->q){
			lp->q = nsock->next;
			nsock->next = nil;
			qunlock(lp);

			if(addr != nil && paddrlen != nil){
				err = getsockaddr(nsock, 1, addr, *paddrlen);
				*paddrlen = err < 0 ? 0 : err;
			}
			return newfd(nsock, FD_CLOEXEC);
		}

		if(sock->omode & O_NONBLOCK){
			err = -EAGAIN;
			break;
		}

		if((err = sleepq(&lp->wq, lp, 1)) < 0)
			break;
	}
	qunlock(lp);

	return err;
}

static int
socketpair(int family, int stype, int protocol, int sv[2])
{
	Socket *sock;
	int p[2];
	int i, fd;

	trace("socketpair(%d, %d, %d, %p)", family, stype, protocol, sv);

	if(family != AF_UNIX)
		return -EAFNOSUPPORT;
	if(pipe(p) < 0)
		return mkerror();
	for(i=0; i<2; i++){
		sock = allocsock(family, stype, protocol);
		sock->fd = p[i];
		sock->connected = 1;
		if((fd = newfd(sock, FD_CLOEXEC)) < 0){
			if(i > 0)
				sys_close(sv[0]);
			close(p[0]);
			close(p[1]);
			return fd;
		}
		sv[i] = fd;
	}
	return 0;
}

static void*
bufprocsock(Socket *sock)
{
	if(sock->bufproc == nil)
		sock->bufproc = newbufproc(sock->fd);
	return sock->bufproc;
}

static int
pollsock(Ufile *file, void *tab)
{
	Socket *sock = (Socket*)file;
	Listenproc *lp;
	Connectproc *cp;

	if(!sock->connected){
		if(lp = sock->listenproc){
			qlock(lp);
			pollwait(file, &lp->wq, tab);
			if(lp->q){
				qunlock(lp);
				return POLLIN;
			}
			qunlock(lp);
		}
		if(cp = sock->connectproc){
			qlock(cp);
			pollwait(file, &cp->wq, tab);
			if(sock->error < 0){
				qunlock(cp);
				return POLLOUT;
			}
			qunlock(cp);
		}
		return 0;
	}

	return pollbufproc(bufprocsock(sock), sock, tab);
}

static int
readsock(Ufile *file, void *buf, int len, vlong)
{
	Socket *sock = (Socket*)file;
	int ret;

	if(!sock->connected)
		return -ENOTCONN;
	if((sock->omode & O_NONBLOCK) || (sock->bufproc != nil)){
		ret = readbufproc(bufprocsock(sock), buf, len, 0, (sock->omode & O_NONBLOCK));
	} else {
		if(notifyme(1))
			return -ERESTART;
		ret = read(sock->fd, buf, len);
		notifyme(0);
		if(ret < 0)
			ret = mkerror();
	}
	return ret;
}

extern int pipewrite(int fd, void *buf, int len);

static int
writesock(Ufile *file, void *buf, int len, vlong)
{
	Socket *sock = (Socket*)file;
	int ret;

	if(!sock->connected)
		return -ENOTCONN;
	if(sock->family == AF_UNIX)
		return pipewrite(sock->fd, buf, len);
	if(notifyme(1))
		return -ERESTART;
	ret = write(sock->fd, buf, len);
	notifyme(0);
	if(ret < 0)
		ret = mkerror();
	return ret;
}

static void
getconf(void *buf)
{
	int len;
	Ipifc *ip, *nip;
	Iplifc *lifc;
	struct ifconf *ifc;
	struct ifreq *ifr, *end;
	struct sockaddr_in *sin;

	ifc = buf;
	ifr = (void *)ifc->ifc_buf;
	end = (struct ifreq *)((uchar *)ifc->ifc_buf + ifc->ifc_len);
	ip = readipifc("/net", nil, -1);

	len = 0;
	for(nip = ip; nip && ifr < end; nip = nip->next){
		if(strcmp(nip->dev, "/dev/null") == 0)
			snprint(ifr->ifr_name, IFNAMSIZ, "lo");
		else
		if(strncmp(nip->dev, "/net/ether", 10) == 0)
			snprint(ifr->ifr_name, IFNAMSIZ, "eth%s", nip->dev+10);

		for(lifc = nip->lifc; lifc && ifr < end; lifc = lifc->next){
			sin = (void *)&ifr->ifr_addr;
			v6tov4((uchar *)&sin->sin_addr.s_addr, (uchar *)lifc->ip);
			sin->sin_family = AF_INET;
			len += sizeof(struct ifreq);
			ifr++;
		}
	}
	ifc->ifc_len = len;
}

static int
ioctlsock(Ufile *file, int cmd, void *arg)
{
	int n;
	Ipifc *ip, *nip;
	uchar addr[4];
	char name[32];
	struct ifreq *ifr;
	struct sockaddr_in *sin;
	Socket *sock = (Socket*)file;

	switch(cmd){
	default:
		return -EINVAL;
	case SIOCGIFADDR:
		ifr = arg;

		ip = readipifc("/net", nil, -1);
		if(strcmp(ifr->ifr_name, "lo") == 0)
			n = -1;
		else
		if(strncmp(ifr->ifr_name, "eth", 3) == 0)
			n = atoi(ifr->ifr_name+3);
		else
			return -ESRCH;

		for(nip = ip; nip; nip = nip->next){
			if(strcmp(nip->dev, "/dev/null") == 0 && n == -1)
				break;
			if(strncmp(nip->dev, "/net/ether", 10) == 0 && atoi(nip->dev+10) == n)
				break;
		}
		if(nip == nil)
			return -ESRCH;

		/* an interface can have multiple addresses, we just pick the first */
		if(nip->lifc == nil)
			return -ESRCH;

		sin = (void *)&ifr->ifr_addr;
		v6tov4((uchar *)&sin->sin_addr.s_addr, (uchar *)nip->lifc->ip);
		sin->sin_family = AF_INET;
		break;
	case SIOCGIFNAME:
		ifr = arg;
		ip = readipifc("/net", nil, -1);
		if(sock->family != AF_INET)
			return -ENODEV;
		for(n = 0, nip = ip; nip ; nip = nip->next, n++)
			if(n >= ifr->ifr_ifindex)
				break;
		if(nip == nil)
			return -ESRCH;
		
		if(strncmp(nip->dev, "/net/ether", 10) == 0)
			snprint(ifr->ifr_name, IFNAMSIZ, "eth%s", nip->dev+10);
		if(strcmp(nip->dev, "/dev/null") == 0)
			snprint(ifr->ifr_name, IFNAMSIZ, "lo");
		break;
	case SIOCGIFCONF:
		getconf(arg);
		break;
	case SIOCGIFFLAGS:
		ifr = arg;
		sin = (void *)&ifr->ifr_addr;

		ip = readipifc("/net", nil, -1);
		for(nip = ip; nip; nip = nip->next){
			if(nip->lifc == nil)
				continue;

			v6tov4((uchar *)addr, (uchar *)nip->lifc->ip);

			if(strcmp(nip->dev, "/dev/null") == 0){
				if(strcmp(ifr->ifr_name, "lo") == 0 || memcmp(&sin->sin_addr.s_addr, addr, 4) == 0){
					ifr->ifr_flags = IFF_UP|IFF_RUNNING|IFF_LOOPBACK;
					break;
				}
			}
			if(strncmp(nip->dev, "/net/ether", 10) == 0){
				snprint(name, sizeof(name), "eth%d", atoi(nip->dev+10));
			  	if(strcmp(ifr->ifr_name, name) == 0 || memcmp(&sin->sin_addr.s_addr, addr, 4) == 0){
					ifr->ifr_flags = IFF_UP|IFF_RUNNING;
					break;
				}
			}
		}
		if(nip == nil)
			return -ENODEV;
		break;
	case FIONREAD:
		{
			int r;

			if(arg == nil)
				return -EINVAL;
			if((r = nreadablebufproc(bufprocsock(sock))) < 0){
				*((int*)arg) = 0;
				return r;
			}
			*((int*)arg) = r;
		}
		break;
	}
	return 0;
}


static int
sendto(Socket *sock, void *data, int len, int flags, struct sockaddr *addr, int addrlen)
{
	if(flags & MSG_OOB)
		return -ENOENT;

	trace("sendto(sock=%p, data=%p, len=%d, flags=%x addr=%p addrlen=%d)",
		sock, data, len, flags, addr, addrlen);

	return writesock(sock, data, len, sock->off);
}

static int
recvfrom(Socket *sock, void *data, int len, int flags, struct sockaddr *addr, int addrlen)
{
	int n, ret;

	trace("recvfrom(sock=%p, data=%p, len=%d, flags=%x addr=%p addrlen=%d)",
		sock, data, len, flags, addr, addrlen);

	if(flags & MSG_OOB)
		return -ENOENT;

	if(flags & MSG_PEEK){
		if(! sock->connected)
			return -ENOTCONN;
		ret = readbufproc(bufprocsock(sock), data, len, 1, 1);
	} else {
		ret = readsock(sock, data, len, sock->off);
	}
	if(addr){
		n = sock->naddr;
		if(addrlen > 0 && n > addrlen)	/* catch lazy programmers */
			n = addrlen;
		memmove(addr, sock->addr, n);
	}
	return ret;
}

enum {
	SOL_SOCKET = 1,

	SO_DEBUG = 1,
	SO_REUSEADDR,
	SO_TYPE,
	SO_ERROR,
};

static int
getoptsock(Socket *sock, int lvl, int opt, char *ov, int *ol)
{
	trace("getoptsock(%p, %d, %d, %p, %p)", sock, lvl, opt, ov, ol);

	switch(lvl){
	default:
	Default:
		return -EINVAL;

	case SOL_SOCKET:
		switch(opt){
		default:
			goto Default;
		case SO_ERROR:
			*ol = sizeof(int);
			*((int*)ov) = sock->error;
			break;
		}
		break;
	}

	return 0;
}

static int
sendmsg(Socket *sock, struct msghdr *msg, int flags)
{
	int i, len, rc;
	char *p, *base;
	
	if(msg->accrights || msg->accrightslen)
		return -ENOENT;

	if(msg->iovlen == 1)
		return sendto(sock, msg->iov->base, msg->iov->len, flags,
			(struct sockaddr *)msg->name, msg->namelen);

	len = 0;
	for(i = 0; i < msg->iovlen; i++)
		len += msg->iov[i].len;
	base = kmalloc(len);
	p = base;
	for(i = 0; i < msg->iovlen; i++){
		memcpy(p, msg->iov[i].base, msg->iov[i].len);
		p += msg->iov[i].len;
	}
	rc = sendto(sock, base, len, flags, (struct sockaddr *)msg->name, msg->namelen);
	free(base);
	return rc;
}

static int
recvmsg(Socket *sock, struct msghdr *msg, int flags)
{
	int i, len, got;
	char *p, *base;

	if(msg->accrights || msg->accrightslen)
		return -ENOENT;

	if(msg->iovlen == 1)
		return recvfrom(sock, msg->iov->base, msg->iov->len, flags,
			(struct sockaddr *)msg->name, msg->namelen);

	len = 0;
	for(i = 0; i < msg->iovlen; i++)
		len += msg->iov[i].len;
	base = kmalloc(len);
	got = recvfrom(sock, base, len, flags, (struct sockaddr *)msg->name, msg->namelen);

	p = base;
	for(i = 0; i < msg->iovlen; i++){
		if(msg->iov[i].len > got)
			msg->iov[i].len = got;
		memcpy(msg->iov[i].base, p, msg->iov[i].len);
		p += msg->iov[i].len;
		got -= msg->iov[i].len;
	}
	free(base);
	return got;
}



enum {
	SYS_SOCKET=1,
	SYS_BIND,
	SYS_CONNECT,
	SYS_LISTEN,
	SYS_ACCEPT,
	SYS_GETSOCKNAME,
	SYS_GETPEERNAME,
	SYS_SOCKETPAIR,
	SYS_SEND,
	SYS_RECV,
	SYS_SENDTO,
	SYS_RECVFROM,
	SYS_SHUTDOWN,
	SYS_SETSOCKOPT,
	SYS_GETSOCKOPT,
	SYS_SENDMSG,
	SYS_RECVMSG,
};

static char *sockfuncs[] = {
	[SYS_SOCKET]		"socket",
	[SYS_BIND]		"bind",
	[SYS_CONNECT]		"connect",
	[SYS_LISTEN]		"listen",
	[SYS_ACCEPT]		"accept",
	[SYS_GETSOCKNAME]	"getsockname",
	[SYS_GETPEERNAME]	"getpeername",
	[SYS_SOCKETPAIR]	"socketpair",
	[SYS_SEND]		"send",
	[SYS_RECV]		"recv",
	[SYS_SENDTO]		"sendto",
	[SYS_RECVFROM]		"recvfrom",
	[SYS_SHUTDOWN]		"shutdown",
	[SYS_SETSOCKOPT]	"setsockopt",
	[SYS_GETSOCKOPT]	"getsockopt",
	[SYS_SENDMSG]		"sendmsg",
	[SYS_RECVMSG]		"recvmsg",
};

int sys_linux_socketcall(int call, int *arg)
{
	Socket *sock;
	int ret;

	if(call > 0 && call <= nelem(sockfuncs) && sockfuncs[call] != nil)
		trace("%s()", sockfuncs[call]);
		
	if(call == SYS_SOCKET)
		return newsock(arg[0], arg[1], arg[2]);

	if(call == SYS_SOCKETPAIR)
		return socketpair(arg[0], arg[1], arg[2], (int*)arg[3]);

	if((sock = (Socket*)fdgetfile(arg[0])) == nil)
		return -EBADF;

	if(sock->dev != SOCKDEV){
		putfile(sock);
		return -ENOTSOCK;
	}

	ret = -1;
	switch(call){
	case SYS_CONNECT:
		ret = connectsock(sock, (void*)arg[1], arg[2]);
		break;
	case SYS_SENDTO:
		ret = sendto(sock, (void*)arg[1], arg[2], arg[3], (void*)arg[4], arg[5]);
		break;
	case SYS_RECVFROM:
		ret = recvfrom(sock, (void*)arg[1], arg[2], arg[3], (void*)arg[4], arg[5]);
		break;
	case SYS_SEND:
		ret = sendto(sock, (void*)arg[1], arg[2], arg[3], nil, 0);
		break;
	case SYS_RECV:
		ret = recvfrom(sock, (void*)arg[1], arg[2], arg[3], nil, 0);
		break;
	case SYS_GETSOCKNAME:
		ret = getsockname(sock, (void*)arg[1], (void*)arg[2]);
		break;
	case SYS_GETPEERNAME:
		ret = getpeername(sock, (void*)arg[1], (void*)arg[2]);
		break;
	case SYS_SHUTDOWN:
		ret = shutdownsock(sock, arg[1]);
		break;
	case SYS_BIND:
		ret = bindsock(sock, (void*)arg[1], arg[2]);
		break;
	case SYS_LISTEN:
		ret = listensock(sock);
		break;
	case SYS_ACCEPT:
		ret = acceptsock(sock, (void*)arg[1], (void*)arg[2]);
		break;
	case SYS_SETSOCKOPT:
		ret = 0;
		break;
	case SYS_GETSOCKOPT:
		ret = getoptsock(sock, (int)arg[1], (int)arg[2], (char*)arg[3], (int*)arg[4]);
		break;
	case SYS_SENDMSG:
		ret = sendmsg(sock, (struct msghdr *)arg[1], (int)arg[2]);
		break;
    	case SYS_RECVMSG:
		ret = recvmsg(sock, (struct msghdr *)arg[1], (int)arg[2]);
		break;
  	default:
		if(call > 0 && call <= nelem(sockfuncs) && sockfuncs[call] != nil)
			kprint("socketcall: %s() %d not implemented\n", sockfuncs[call], call);
		else
			kprint("socketcall: ?() %d not implemented\n", call);
	}

	putfile(sock);
	return ret;
}

static void
fillstat(Ustat *s)
{
	s->mode = 0666 | S_IFSOCK;
	s->uid = current->uid;
	s->gid = current->gid;
	s->size = 0;
}

static int
fstatsock(Ufile *, Ustat *s)
{
	fillstat(s);
	return 0;
};

static Udev sockdev = 
{
	.name = "sock",
	.read = readsock,
	.write = writesock,
	.poll = pollsock,
	.close = closesock,
	.ioctl = ioctlsock,
	.fstat = fstatsock,
};

void sockdevinit(void)
{
	devtab[SOCKDEV] = &sockdev;
}

