/* ls.c - list files */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <utime.h>
#include <pwd.h>
#include <grp.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>


#define CHUNK		50

typedef struct {
	char *name;
	char *prefix;
	struct stat sb;
	char *user;
	char *group;
} Dir;

static int Fflag = 0;
static int dflag = 0;
static int qflag = 0;
static int lflag = 0;
static int nflag = 0;
static int pflag = 0;
static int rflag = 0;
static int sflag = 0;
static int tflag = 0;
static int uflag = 0;

static int ndir = 0;
static char *argv0;
static time_t clk;
static FILE *out;
static Dir *dirbuf;

extern int main(int argc, char *argv[]);
extern int ls(char *s, int multi);
extern void output(void);
extern void dowidths(Dir *dt);
extern char *fileflag(int m);
extern char *modestr(int m);
extern char *asciitime(time_t l);
extern void addent(char *prefix, char *name, struct stat *sbp);
extern void format(Dir *dt, char *name);
extern int dircmp(const void *x, void const *y);
extern int usage(void);

int	dwidth;			/* max width of device id in hex */
int	swidth;			/* max width of -s size */
int	qwidth;			/* max width of -q version */
int	vwidth;			/* max width of dev */
int	uwidth;			/* max width of userid */
int	mwidth;			/* max width of muid */
int	lwidth;			/* max width of length */
int	gwidth;			/* max width of groupid */

int
main(int argc, char *argv[])
{
	int o, n;

	n = 0;
	argv0 = argv[0];
	while ((o = getopt(argc, argv, "Fdlnqprstu")) != EOF)
		switch (o){
		case 'F': Fflag++; break;	// flag dirs and executables
		case 'd': dflag++; break;	// dont open dirs
		case 'l': lflag++; break;	// long format list
		case 'n': nflag++; break;	// no sort
		case 'q': qflag++; break;	// list qid (Inode simulation)
		case 'p': pflag++; break;	// do not prepend paths to filenames
		case 'r': rflag++; break;	// reverse sense of sort
		case 's': sflag++; break;	// print size in Kb and name
		case 't': tflag++; break;	// sort by & print creaion time
		case 'u': uflag++; break;	// sort by & print access time
		default: return usage();
		}

	if(lflag){
		tzset();
		time(&clk);
	}

	if(optind >= argc)
		ls(".", 0);
	else
		for (; optind < argc; optind++)
			ls(argv[optind], 1);

	output();
	return 0;
}

int
ls(char *s, int multi)
{
	DIR *d;
	int rc;
	struct stat sb;
	struct dirent *de;
	char *p, file[FILENAME_MAX];

	rc = 0;
	out = stdout;

	while((p = strrchr(s, '/')) && p != s && p[1] == 0)
		*p = 0;

	if(stat(s, &sb) == -1){
		fprintf(stderr, "%s: %s - %s\n", argv0, s, strerror(errno));
		return 1;
	}

	if((sb.st_mode & S_IFDIR) && dflag == 0){
		output();
		if((d = opendir(s)) == NULL){
			fprintf(stderr, "%s:  %s - %s\n", argv0, s, strerror(errno));
			return 1;
		}
		while((de = readdir(d)) != NULL){
			if(de->d_name[0] == '.' &&
			   (de->d_name[1] == '\0' ||
			   (de->d_name[1] == '.' &&
			    de->d_name[2] == '\0')))
				continue;

			strcpy(file, s);
			if(strcmp(s, "/") != 0)	/* root is special */
				strcat(file, "/");
			strcat(file, de->d_name);

			if(stat(file, &sb) == -1){
				fprintf(stderr, "%s: %s - %s\n", argv0, file, strerror(errno));
				return 1;
			}

			addent(((multi)? s : NULL), de->d_name, &sb);
                }
                output();
	}
	else{
                addent(NULL, s, &sb);
	}

	return rc;
}

void
output(void)
{
	int i;
	char *s, buf[4096];

	if(! nflag)
		qsort(dirbuf, ndir, sizeof(Dir), dircmp);
	for(i = 0; i < ndir; i++)
		dowidths(&dirbuf[i]);
	for(i = 0; i < ndir; i++){
		s = dirbuf[i].prefix;
		if(!pflag && s != NULL){
			if(strcmp(s, "/") == 0)  /* root dir is a special case */
				s = "";
			sprintf(buf, "%s/%s", s, dirbuf[i].name);
			format(&dirbuf[i], buf);
		}
		else
			format(&dirbuf[i], dirbuf[i].name);
	}
	fflush(out);
	ndir = 0;
}

void
dowidths(Dir *dt)
{
	int n;
	char buf[256], num[16];
	struct group *gr;
	struct passwd *pw;

	if(sflag){
		n = sprintf(buf, "%lu", (dt->sb.st_size+1023)/1024);
		if(n > swidth)
			swidth = n;
	}
	if(qflag){
		n = sprintf(buf, "%lu", dt->sb.st_ino);
		if(n > qwidth)
			qwidth = n;
	}
	if(lflag){
		if((pw = getpwuid(dt->sb.st_uid)) == NULL){
			snprintf(num, sizeof(num), "(%d)", dt->sb.st_uid);
			dt->user = strdup(num);
		}
		else
			dt->user = strdup(pw->pw_name);
		if((gr = getgrgid(dt->sb.st_gid)) == NULL){
			snprintf(num, sizeof(num), "(%d)", dt->sb.st_uid);
			dt->group = strdup(num);
		}
		else
			dt->group = strdup(gr->gr_name);

		n = sprintf(buf, "%u", major(dt->sb.st_dev));
		if(n > dwidth)
			dwidth = n;
		n = sprintf(buf, "%u", minor(dt->sb.st_dev));
		if(n > vwidth)
			vwidth = n;
		n = sprintf(buf, "%s", dt->user);
		if(n > uwidth)
			uwidth = n;
		n = sprintf(buf, "%s", dt->group);
		if(n > gwidth)
			gwidth = n;
		n = sprintf(buf, "%lu", dt->sb.st_size);
		if(n > lwidth)
			lwidth = n;
	}
}

char *
fileflag(int m)
{
        if(Fflag == 0)
                return("");
        if(m & S_IFDIR)
                return("/");
        if(m & 0111)
                return("*");

        return "";
}


char *
modestr(int m)
{
	static char buf[32];

	strcpy(buf, "----------");

	if((m & S_IFDIR) == S_IFDIR)
		buf[0] = 'd';
	if((m & S_IFCHR) == S_IFCHR)
		buf[0] = 'c';
	if((m & S_IFBLK) == S_IFBLK)
		buf[0] = 'b';
	if((m & S_IFLNK) == S_IFLNK)
		buf[0] = 'l';
	if((m & S_IFSOCK) == S_IFSOCK)
		buf[0] = 's';
	if((m & S_IFIFO) == S_IFIFO)
		buf[0] = 'f';

	if(m & S_IREAD)
		buf[1] = 'r';
	if(m & S_IWRITE)
		buf[2] = 'w';
	if(m & S_IEXEC)
		buf[3] = 'x';

	m <<= 3;

	if(m & S_IREAD)
		buf[4] = 'r';
	if(m & S_IWRITE)
		buf[5] = 'w';
	if(m & S_IEXEC)
		buf[6] = 'x';

	m <<= 3;

	if(m & S_IREAD)
		buf[7] = 'r';
	if(m & S_IWRITE)
		buf[8] = 'w';
	if(m & S_IEXEC)
		buf[9] = 'x';

	return buf;
}

char *
asciitime(time_t l)
{
        char *t;
        static char buf[32];

        t = ctime(&l);

        /* 6 months in the past or a day in the future */
        if(l < clk - 180L*24*60*60 || clk + 24L*60*60 < l){
                memmove(buf, t+4, 7);           /* month and day */
                memmove(buf+7, t+19, 5);        /* year */
        }else
                memmove(buf, t+4, 12);          /* skip day of week */

        buf[12] = 0;
        return buf;
}
   

void
addent(char *prefix, char *name, struct stat *sbp)
{
        static int alloc = 0;

	if(ndir >= alloc){
		alloc += CHUNK;
		if((dirbuf = realloc(dirbuf, alloc * sizeof(Dir))) == NULL){
			fprintf(stderr, "ls: realloc failed\n");
			exit(1);
		}
	}

	memmove(&dirbuf[ndir].sb, sbp, sizeof(struct stat));
	dirbuf[ndir].name = strdup(name);
	dirbuf[ndir].prefix = prefix;
	ndir++;
}

void
format(Dir *dt, char *name)
{
       if(sflag)
                fprintf(out, "%4lu ", (dt->sb.st_size + 1023L)/ 1024L);

        if(qflag)
                fprintf(out, "%*lu ", qwidth, dt->sb.st_ino);

        if(lflag)
		fprintf(out, "%s %*u %*u %*s %*s %*lu %s ",
			modestr(dt->sb.st_mode),
                       	dwidth, major(dt->sb.st_dev), 
			vwidth, minor(dt->sb.st_dev),
			-uwidth, dt->user,
			-gwidth, dt->group, 
                        lwidth, dt->sb.st_size,
                        asciitime(uflag? dt->sb.st_atime : dt->sb.st_mtime));

 	fprintf(out, "%s%s\n", name, fileflag(dt->sb.st_mode));
}

int
dircmp(const void *x, void const *y)
{
	long rc;
	int rev;
	Dir *a = (Dir *)x;
	Dir *b = (Dir *)y;


	if(rflag)
		rev = -1;
	else
		rev = 1;

	if(tflag){
		if(uflag)
			return (b->sb.st_atime-a->sb.st_atime)*rev;
		else
			return (b->sb.st_mtime-a->sb.st_mtime)*rev;
	}
	else
		return strcmp(a->name, b->name)*rev;

	return (int)rc;
}

int
usage()
{
	fprintf(stderr, "usage: %s [-Fdlnprstu] [file ...]\n", argv0);
	fprintf(stderr, "-F     flag directories and executables\n");
	fprintf(stderr, "-d     don't decend into directories\n");
	fprintf(stderr, "-l     long listing\n");
	fprintf(stderr, "-n     don't sort\n");
	fprintf(stderr, "-p     don't append file's path\n");
	fprintf(stderr, "-q     print the qid of the file\n");
	fprintf(stderr, "-r     reverse sense of sort\n");
	fprintf(stderr, "-s     prefix files by their size (in Kbytes)\n");
	fprintf(stderr, "-t     sort by modification time\n");
	fprintf(stderr, "-u     sort/list access time\n");
	exit(1);
}
