#include <u.h>
#include <libc.h>
#include <ureg.h>
#include "dat.h"
#include "fns.h"
#include "linux.h"

#pragma profile off

typedef struct {
	char *msg;
	int chatty;
	void (*func)(int, int);
	int sig;
	int code;
	int hash;
} Dispatch;

static void intr(int, int);
static void sig(int, int);
static void gpf(int, int);

enum { X86_INT80 = 0x80cd };

/*
 * NB: table roughly sorted in inverse probability of occurance
 */
static Dispatch disptab[] = {
	{ "sys: trap: general protection violation pc",	0, gpf, SIGSEGV, 0 },
	{ "interrupt",					1, intr, SIGINT, 0 },
	{ "kill",					1, intr, SIGQUIT, 0 },
	{ "sys: write on closed pipe pc",		1, sig, SIGPIPE, 0 },
	{ "sys: trap: divide error pc",			1, sig, SIGFPE, FPE_INTDIV },
	{ "sys: trap: overflow pc",			1, sig, SIGFPE, FPE_INTOVF },
	{ "sys: fp: unknown exception fppc",		1, sig, SIGFPE, 0 },
	{ "sys: fp: denormalized operand fppc",		1, sig, SIGFPE, FPE_FLTRES },	/* FIXME: FPE_FLTRES is a guess */
	{ "sys: fp: division by zero fppc",		1, sig, SIGFPE, FPE_FLTDIV },
	{ "sys: fp: numeric overflow fppc",		1, sig, SIGFPE, FPE_FLTOVF },
	{ "sys: fp: numeric underflow fppc",		1, sig, SIGFPE, FPE_FLTUND },
	{ "sys: fp: precision loss fppc",		1, sig, SIGFPE, FPE_FLTRES },
	{ "sys: fp: stack overflow fppc",		2, sig, SIGFPE, 0 },
	{ "sys: fp: stack underflow fppc",		2, sig, SIGFPE, 0 },
	{ "sys: fp: invalid operation fppc",		2, sig, SIGFPE, FPE_FLTINV },

	{ "sys: trap: alignment check",			2, sig, SIGBUS, BUS_ADRALN },

	{ "sys: trap: fault read addr",			2, sig, SIGSEGV, SEGV_ACCERR },
	{ "sys: trap: fault write addr",		2, sig, SIGSEGV, SEGV_ACCERR },
	{ "sys: trap: invalid opcode pc",		2, sig, SIGILL, ILL_ILLOPC },

	{ "sys: sse error", 				2, sig, SIGILL, ILL_COPROC },

	/* see /sys/src/ape/lib/ap/plan9/signal.c */
	{ "sys: trap: arithmetic overflow",		1, sig, SIGFPE, FPE_INTOVF },	/*  old name for "trap overflow" above ? */
	{ "sys: trap: bus error",			2, sig, SIGBUS, BUS_OBJERR },
	{ "sys: trap: address error",			2, sig, SIGSEGV, BUS_ADRERR },
	{ "sys: trap: TLB",				2, sig, SIGSEGV, 0 },
	{ "sys: trap: illegal instruction",		2, sig, SIGILL, ILL_ILLOPC },
	{ "sys: trap: reserved instruction",		2, sig, SIGILL, ILL_PRVOPC },
	{ "sys: trap: reserved",			2, sig, SIGILL, 0 },
};

void
retuser(void)
{
	Uproc *p;
	Ureg *u;

	p = current;
	u = p->ureg;
	p->ureg = nil;
	if(p->innote == 0)
		jumpureg(u);
	p->innote = 0;
	noted(NCONT);
}

static void
intr(int sig, int code)
{
	Uproc *p;
	Usiginfo si;

	p = current;
	if(p->notified){
		p->notified = 0;
		return;
	}
	memset(&si, 0, sizeof(si));
	si.signo = sig;
	si.code = code;
	si.fault.addr = (void *)p->ureg->pc;
	sendsignal(p, &si, 0);
}

static void
sig(int sig, int code)
{
	Uproc *p;
	Usiginfo si;

	p = current;
	memset(&si, 0, sizeof(si));
	si.signo = sig;
	si.code = code;
	si.fault.addr = (void *)p->ureg->pc;
	sendsignal(p, &si, 0);
}

static void
gpf(int sig, int code)
{
	Uproc *p;
	Usiginfo si;
	unsigned *pc;

	p = current;
	pc = (void *)p->ureg->pc;
	if((*pc & 0xffff) == X86_INT80){
		linuxcall();
		return;
	}
	memset(&si, 0, sizeof(si));
	si.signo = sig;
	si.code = code;
	si.fault.addr = (void *)p->ureg->pc;
	sendsignal(p, &si, 0);
}

/*
 * as sum is strong enough and speed is of the essence
 */
static ulong
msghash(char *msg)
{
	ulong s;

	s = 0;
	while(*msg && *msg != '=')
		s += *msg++;
	return s;
}

static void
handletrap(void *v, char *msg)
{
	ulong hash;
	Dispatch *d;
	Uproc *p;
	uchar *pc;

	
	p = current;
	p->innote = 1;
	p->ureg = v;
	pc = (void *)p->ureg->pc;

	if((*pc & 0xffff) == X86_INT80){	/* This is a rather vile shortcut, for speed */
		linuxcall();
	}
	else{

		hash = msghash(msg);
		for(d = disptab; d < &disptab[nelem(disptab)]; d++)
			if(d->hash == hash)
				break;

		if(d)
			d->func(d->sig, d->code);

		if(d == nil || d->chatty > 1){
			trace("handletrap: %s", msg);
			if(p && p->comm)
				kprint("%s ", p->comm);
			kprint("%s ", msg);
			if(pc > (uchar *)0xffff)
				kprint("*pc=%02x,%02x,%02x,%02x ", pc[0], pc[1], pc[2], pc[3]);
			kprint("\n");

		}

		if(d == nil || d->chatty >= 2)
			noted(NDFLT);
	}

	if(p->traceproc)
		p->traceproc(p->tracearg);

	handlesignals();
	retuser();
}

#pragma profile on


void
inittrap(void)
{
	ulong f;
	Dispatch *d;

	/* disable FPU faults */
	f = getfcr();
	f &= ~(FPINEX|FPOVFL|FPUNFL|FPZDIV|FPINVAL);
	setfcr(f);

	for(d = disptab; d < &disptab[nelem(disptab)]; d++)
		d->hash = msghash(d->msg);

	notify(handletrap);
}
